
# THIS FILE IS GENERATED FROM linkahead SETUP.PY
#
short_version = '0.15.1'
version = '0.15.1'
full_version = '0.15.1'
git_revision = '86ffb96add53e78c0de8f5024a072d974f958adb'
release = True

if not release:
    version = full_version



# -*- coding: utf-8 -*-
#
# This file is a part of the LinkAhead Project.
#
# Copyright (C) 2023 Henrik tom Wörden <h.tomwoerden@indiscale.com>
# Copyright (C) 2023 IndiScale GmbH <info@indiscale.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

"""Convenience functions to retrieve a specific entity."""

from typing import Union, Optional

from ..common.models import Entity, execute_query
from .escape import escape_squoted_text


def get_entity_by_name(name: str, role: Optional[str] = None) -> Entity:
    """Return the result of a unique query that uses the name to find the correct entity.

Submits the query "FIND {role} WITH name='{name}'".

Parameters
----------

role: str, optional
  The role for the query, defaults to ``ENTITY``.
    """
    name = escape_squoted_text(name)
    if role is None:
        role = "ENTITY"
    # type hint can be ignored, it's a unique query, so never Container or int
    return execute_query(f"FIND {role} WITH name='{name}'", unique=True)  # type: ignore


def get_entity_by_id(eid: Union[str, int], role: Optional[str] = None) -> Entity:
    """Return the result of a unique query that uses the id to find the correct entity.

Submits the query "FIND {role} WITH id='{eid}'".

Parameters
----------

role: str, optional
  The role for the query, defaults to ``ENTITY``.
    """
    if role is None:
        role = "ENTITY"
    # type hint can be ignored, it's a unique query
    return execute_query(f"FIND {role} WITH id='{eid}'", unique=True)  # type: ignore


def get_entity_by_path(path: str) -> Entity:
    """Return the result of a unique query that uses the path to find the correct file.

Submits the query "FIND {role} WHICH IS STORED AT '{path}'".

Parameters
----------

role: str, optional
  The role for the query, defaults to ``ENTITY``.
    """
    # type hint can be ignored, it's a unique query
    return execute_query(f"FIND FILE WHICH IS STORED AT '{path}'", unique=True)  # type: ignore



# -*- coding: utf-8 -*-
#
# ** header v3.0
# This file is a part of the LinkAhead Project.
#
# Copyright (C) 2018 Research Group Biomedical Physics,
# Max-Planck-Institute for Dynamics and Self-Organization Göttingen
# Copyright (C) 2020 Timm Fitschen <t.fitschen@indiscale.com>
# Copyright (C) 2020 IndiScale GmbH <info@indiscale.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
# ** end header
#
from __future__ import annotations
"""Utility functions for server and user administration."""

import random
import re
import string

from lxml import etree

from ..connection.connection import get_connection
from ..exceptions import (EntityDoesNotExistError, HTTPClientError,
                          HTTPForbiddenError, HTTPResourceNotFoundError,
                          ServerConfigurationException)
from .utils import xml2str

from typing import Optional, TYPE_CHECKING, Union
if TYPE_CHECKING:
    from ..common.models import Entity


def set_server_property(key: str, value: str):
    """set_server_property.

    Set a server property.

    Parameters
    ----------
    key : str
        The name of the server property.
    value : str
        The value of the server property.


    Returns
    -------
    None
    """
    con = get_connection()
    try:
        con._form_data_request(method="POST", path="_server_properties",
                               params={key: value}).read()
    except EntityDoesNotExistError:
        raise ServerConfigurationException(
            "Debug mode in server is probably disabled.") from None


def get_server_properties() -> dict[str, Optional[str]]:
    """get_server_properties.

    Get all server properties as a dict.

    Returns
    -------
    dict
        The server properties.
    """
    con = get_connection()
    try:
        body = con._http_request(
            method="GET", path="_server_properties")
    except EntityDoesNotExistError:
        raise ServerConfigurationException(
            "Debug mode in server is probably disabled.") from None

    xml = etree.parse(body)
    props: dict[str, Optional[str]] = dict()

    for elem in xml.getroot():
        props[elem.tag] = elem.text

    return props


def get_server_property(key: str) -> Optional[str]:
    """get_server_property.

    Get a server property.

    Parameters
    ----------
    key : str
        The name of the server property

    Returns
    -------
    value : str
        The string value of the server property.

    Raises
    ------
    KeyError
        If the server property is no defined.
    """

    return get_server_properties()[key]


def generate_password(length: int):
    """Create a random password that fulfills the security requirements

    Parameters
    ----------
    length : int
        Length of the generated password.  Has to be greater than 7.

    Returns
    -------
    password : string
        Generated random password of the given length

    Raises
    ------
    ValueError:
        If the length is less than 8.
    """
    minimum_password_length = 8
    if length < minimum_password_length:
        raise ValueError("LinkAhead passwords have to be at least {} characters.".format(
            minimum_password_length))
    sample_letters = string.ascii_letters + string.digits + "!#$%*+-/:;?_"
    password = ''.join((random.choice(sample_letters) for i in range(length)))

    while not re.match(r"(?=.*[A-Z])(?=.*[a-z])(?=.*\d)(?=.*[\W_]).{8,}",
                       password):
        password = ''.join((random.choice(sample_letters)
                            for i in range(length)))

    return password


def _retrieve_user(name: str, realm: Optional[str] = None, **kwargs):
    con = get_connection()
    try:
        return con._http_request(method="GET", path="User/" + (realm + "/" + name if realm is not None else name), **kwargs).read()
    except HTTPForbiddenError as e:
        e.msg = "You are not permitted to retrieve this user."
        raise
    except HTTPResourceNotFoundError as e:
        e.msg = "User does not exist."
        raise


def _delete_user(name: str, **kwargs):
    con = get_connection()
    try:
        return con._http_request(method="DELETE", path="User/" + name, **kwargs).read()
    except HTTPForbiddenError as e:
        e.msg = "You are not permitted to delete this user."
        raise
    except HTTPResourceNotFoundError as e:
        e.msg = "User does not exist."
        raise


def _update_user(name: str,
                 realm: Optional[str] = None,
                 password: Optional[str] = None,
                 status: Optional[str] = None,
                 email: Optional[str] = None,
                 entity: Optional[Entity] = None, **kwargs):
    con = get_connection()
    params: dict[str, Optional[str]] = {}

    if password is not None:
        params["password"] = password

    if status is not None:
        params["status"] = status

    if email is not None:
        params["email"] = email

    if entity is not None:
        params["entity"] = str(entity)
    try:
        return con.put_form_data(entity_uri_segment="User/" + (realm + "/" + name if realm is not None else name), params=params, **kwargs).read()
    except HTTPResourceNotFoundError as e:
        e.msg = "User does not exist."
        raise e
    except HTTPForbiddenError as e:
        e.msg = "You are not permitted to update this user."
        raise e
    except HTTPClientError as e:
        for elem in etree.fromstring(e.body):
            if elem.tag == "Error":
                e.msg = elem.get("description")
        raise


def _insert_user(name: str,
                 password: Optional[str] = None,
                 status: Optional[str] = None,
                 email: Optional[str] = None,
                 entity: Optional[Entity] = None, **kwargs):
    con = get_connection()
    params: dict[str, Union[str, Entity]] = {"username": name}

    if password is not None:
        params["password"] = password

    if status is not None:
        params["status"] = status

    if email is not None:
        params["email"] = email

    if entity is not None:
        params["entity"] = entity
    try:
        return con.post_form_data(entity_uri_segment="User", params=params, **kwargs).read()
    except HTTPForbiddenError as e:
        e.msg = "You are not permitted to insert a new user."
        raise e
    except HTTPClientError as e:
        for elem in etree.fromstring(e.body):
            if elem.tag == "Error":
                e.msg = elem.get("description")
        raise e


def _insert_role(name, description, **kwargs):
    con = get_connection()
    try:
        return con.post_form_data(entity_uri_segment="Role", params={"role_name": name, "role_description": description}, **kwargs).read()
    except HTTPForbiddenError as e:
        e.msg = "You are not permitted to insert a new role."
        raise
    except HTTPClientError as e:
        if e.status == 409:
            e.msg = "Role name is already in use. Choose a different name."
        raise


def _update_role(name, description, **kwargs):
    con = get_connection()
    try:
        return con.put_form_data(entity_uri_segment="Role/" + name, params={"role_description": description}, **kwargs).read()
    except HTTPForbiddenError as e:
        e.msg = "You are not permitted to update this role."
        raise
    except HTTPResourceNotFoundError as e:
        e.msg = "Role does not exist."
        raise


def _retrieve_role(name, **kwargs):
    con = get_connection()
    try:
        return con._http_request(method="GET", path="Role/" + name, **kwargs).read()
    except HTTPForbiddenError as e:
        e.msg = "You are not permitted to retrieve this role."
        raise
    except HTTPResourceNotFoundError as e:
        e.msg = "Role does not exist."
        raise


def _delete_role(name, **kwargs):
    con = get_connection()
    try:
        return con._http_request(method="DELETE", path="Role/" + name, **kwargs).read()
    except HTTPForbiddenError as e:
        e.msg = "You are not permitted to delete this role."
        raise
    except HTTPResourceNotFoundError as e:
        e.msg = "Role does not exist."
        raise


def _set_roles(username, roles, realm=None, **kwargs):
    xml = etree.Element("Roles")

    for r in roles:
        xml.append(etree.Element("Role", name=r))

    body = xml2str(xml)
    con = get_connection()
    try:
        body = con._http_request(method="PUT", path="UserRoles/" + (realm + "/" +
                                                                    username if realm is not None else username), body=body, **kwargs).read()
    except HTTPForbiddenError as e:
        e.msg = "You are not permitted to set this user's roles."
        raise
    except HTTPResourceNotFoundError as e:
        e.msg = "User does not exist."
        raise
    except HTTPClientError as e:
        if e.status == 409:
            e.msg = "Role does not exist."
        raise
    ret = set()

    for r in etree.fromstring(body)[0]:
        if r.tag == "Role":
            ret.add(r.get("name"))

    return ret


def _get_roles(username, realm=None, **kwargs):
    con = get_connection()
    try:
        body = con._http_request(method="GET", path="UserRoles/" + (
            realm + "/" + username if realm is not None else username), **kwargs).read()
    except HTTPForbiddenError as e:
        e.msg = "You are not permitted to retrieve this user's roles."
        raise
    except HTTPResourceNotFoundError as e:
        e.msg = "User does not exist."
        raise
    ret = set()

    for r in etree.fromstring(body).xpath('/Response/Roles')[0]:
        if r.tag == "Role":
            ret.add(r.get("name"))

    return ret


def _set_permissions(role, permission_rules, **kwargs):
    """Set permissions for a role.

Parameters
----------

role : str
    The role for which the permissions are set.

permission_rules : iterable<PermissionRule>
    An iterable with PermissionRule objects.

**kwargs :
    Additional arguments which are passed to the HTTP request.

Returns
-------
    None
    """
    xml = etree.Element("PermissionRules")

    for p in permission_rules:
        xml.append(p._to_xml())

    body = xml2str(xml)
    con = get_connection()
    try:
        return con._http_request(method="PUT", path="PermissionRules/" + role, body=body, **kwargs).read()
    except HTTPForbiddenError as e:
        e.msg = "You are not permitted to set this role's permissions."
        raise
    except HTTPResourceNotFoundError as e:
        e.msg = "Role does not exist."
        raise


def _get_permissions(role, **kwargs):
    con = get_connection()
    try:
        return PermissionRule._parse_body(con._http_request(method="GET", path="PermissionRules/" + role, **kwargs).read())
    except HTTPForbiddenError as e:
        e.msg = "You are not permitted to retrieve this role's permissions."
        raise
    except HTTPResourceNotFoundError as e:
        e.msg = "Role does not exist."
        raise


class PermissionRule():
    """Permission rules.

Parameters
----------
action : str
    Either "grant" or "deny"

permission : str
    For example ``RETRIEVE:*``.

priority : bool, optional
    Whether the priority shall be set, defaults is False.
    """

    @staticmethod
    def _parse_boolean(bstr) -> bool:
        return str(bstr) in ["True", "true", "TRUE", "yes"]

    def __init__(self, action: str, permission: str, priority: bool = False):
        self._action = action
        self._permission = permission
        self._priority = PermissionRule._parse_boolean(priority)

    def _to_xml(self) -> etree._Element:
        xml = etree.Element(self._action)
        xml.set("permission", self._permission)

        if self._priority is True:
            xml.set("priority", "true")

        return xml

    @staticmethod
    def _parse_element(elem: etree._Element):
        permission = elem.get("permission")
        if permission is None:
            raise ValueError(f"Permission is missing in PermissionRule xml: {elem}")
        priority = PermissionRule._parse_boolean(elem.get("priority"))
        return PermissionRule(elem.tag, permission, priority if priority is not None else False)

    @staticmethod
    def _parse_body(body: str):
        xml = etree.fromstring(body)
        ret = set()

        for c in xml:
            if c.tag in ["Grant", "Deny"]:
                ret.add(PermissionRule._parse_element(c))

        return ret

    def __str__(self):
        return str(self._action) + "(" + str(self._permission) + ")" + \
            ("P" if self._priority is True else "")

    def __repr__(self):
        return str(self)

    def __hash__(self):
        return hash(str(self).lower())

    def __eq__(self, other):
        return str(other).lower() == str(self).lower()



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# ** header v3.0
# This file is a part of the LinkAhead Project.
#
# Copyright (C) 2018 Research Group Biomedical Physics,
# Max-Planck-Institute for Dynamics and Self-Organization Göttingen
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
# ** end header
#

"""requests the server to execute checkFileSystemConsistency job."""

import sys
import linkahead as db

from argparse import ArgumentParser
from argparse import RawDescriptionHelpFormatter

__version__ = 0.1
__date__ = '2016-08-31'
__updated__ = '2016-09-01'


class CLIError(Exception):
    """Generic exception to raise and log different fatal errors."""

    def __init__(self, msg):
        super(CLIError).__init__(type(self))
        self.msg = "E: %s" % msg

    def __str__(self):
        return self.msg

    def __unicode__(self):
        return self.msg


def runCheck(timeout, location):
    """ Request the LinkAhead server to check the file system for consistency.

    location == None means that the whole file system is being checked.
    Otherwise only a the directory tree under location is being checked.
    """

    if (timeout is not None):
        db.get_config().set("Connection", "timeout", str(100 + int(timeout)))
    files = db.Container().retrieve(
        unique=False, raise_exception_on_error=False, flags={
            "fileStorageConsistency": (
                "-t " + str(timeout) if timeout else "") + (
                location if location else ""), })
    return files


def main(argv=None):
    """Command line options."""

    if argv is None:
        argv = sys.argv
    else:
        sys.argv.extend(argv)

    # program_name = os.path.basename(sys.argv[0])
    program_version = "v%s" % __version__
    program_build_date = str(__updated__)
    program_version_message = '%%(prog)s %s (%s)' % (
        program_version, program_build_date)
    program_license = '''

  Copyright 2016 BMPG. All rights reserved.

  Distributed on an "AS IS" basis without warranties
  or conditions of any kind, either express or implied.

USAGE
'''

    # Setup argument parser
    parser = ArgumentParser(description=program_license,
                            formatter_class=RawDescriptionHelpFormatter)
    parser.add_argument(
        "-v",
        "--verbose",
        dest="verbose",
        action="count",
        help="set verbosity level [default: %(default)s]",
        default=0)
    parser.add_argument('-V', '--version', action='version',
                        version=program_version_message)
    parser.add_argument(
        '-t',
        '--timeout',
        dest="timeout",
        help="timeout in seconds for the database requests. [default: %(default)s]",
        metavar="TIMEOUT",
        default="200")
    parser.add_argument('location')

    # Process arguments
    args = parser.parse_args()
    global VERBOSITY

    VERBOSITY = args.verbose
    TIMEOUT = args.timeout

    print(runCheck(TIMEOUT, args.location).messages)

    return 0


if __name__ == "__main__":
    sys.exit(main())



# -*- coding: utf-8 -*-
#
# This file is a part of the LinkAhead Project.
#
# Copyright (C) 2018 Research Group Biomedical Physics,
# Max-Planck-Institute for Dynamics and Self-Organization Göttingen
# Copyright (C) 2020 Timm Fitschen <t.fitschen@indiscale.com>
# Copyright (C) 2020-2022 IndiScale GmbH <info@indiscale.com>
# Copyright (C) 2022 Alexander Schlemmer <alexander.schlemmer@ds.mpg.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
# ** end header
#
# type: ignore
"""
A high level API for accessing LinkAhead entities from within python.

This is refactored from apiutils.
"""

import warnings
from dataclasses import dataclass, fields
from datetime import datetime
from typing import Any, Dict, List, Optional, Union

import yaml
from dateutil import parser

import linkahead as db
from .apiutils import create_flat_list, get_type_of_entity_with
from .common.datatype import (BOOLEAN, DATETIME, DOUBLE, FILE, INTEGER,
                              REFERENCE, TEXT, get_list_datatype,
                              is_list_datatype, is_reference)

warnings.warn("""EXPERIMENTAL! The high_level_api module is experimental and may be changed or
removed in the future. Its purpose is to give an impression on how the Python client user interface
might be changed.""")


def standard_type_for_high_level_type(high_level_record: "CaosDBPythonEntity",
                                      return_string: bool = False):
    """
    For a given CaosDBPythonEntity either return the corresponding
    class in the standard CaosDB API or - if return_string is True - return
    the role as a string.
    """
    if type(high_level_record) == CaosDBPythonRecord:
        if not return_string:
            return db.Record
        return "Record"
    elif type(high_level_record) == CaosDBPythonFile:
        if not return_string:
            return db.File
        return "File"
    elif type(high_level_record) == CaosDBPythonProperty:
        if not return_string:
            return db.Property
        return "Property"
    elif type(high_level_record) == CaosDBPythonRecordType:
        if not return_string:
            return db.RecordType
        return "RecordType"
    elif type(high_level_record) == CaosDBPythonEntity:
        if not return_string:
            return db.Entity
        return "Entity"
    raise RuntimeError("Incompatible type.")


def high_level_type_for_role(role: str):
    if role == "Record":
        return CaosDBPythonRecord
    if role == "File":
        return CaosDBPythonFile
    if role == "Property":
        return CaosDBPythonProperty
    if role == "RecordType":
        return CaosDBPythonRecordType
    if role == "Entity":
        return CaosDBPythonEntity
    raise RuntimeError("Unknown role.")


def high_level_type_for_standard_type(standard_record: db.Entity):
    if not isinstance(standard_record, db.Entity):
        raise ValueError()
    role = standard_record.role
    if role == "Record" or type(standard_record) == db.Record:
        return CaosDBPythonRecord
    elif role == "File" or type(standard_record) == db.File:
        return CaosDBPythonFile
    elif role == "Property" or type(standard_record) == db.Property:
        return CaosDBPythonProperty
    elif role == "RecordType" or type(standard_record) == db.RecordType:
        return CaosDBPythonRecordType
    elif role == "Entity" or type(standard_record) == db.Entity:
        return CaosDBPythonEntity
    raise RuntimeError("Incompatible type.")


@dataclass
class CaosDBPropertyMetaData:
    # name is already the name of the attribute
    unit: Optional[str] = None
    datatype: Optional[str] = None
    description: Optional[str] = None
    id: Optional[int] = None
    importance: Optional[str] = None


class CaosDBPythonUnresolved:
    pass


@dataclass
class CaosDBPythonUnresolvedParent(CaosDBPythonUnresolved):
    """
    Parents can be either given by name or by ID.

    When resolved, both fields should be set.
    """

    id: Optional[int] = None
    name: Optional[str] = None


@dataclass
class CaosDBPythonUnresolvedReference(CaosDBPythonUnresolved):

    def __init__(self, id=None):
        self.id = id


class CaosDBPythonEntity(object):

    def __init__(self):
        """
        Initialize a new CaosDBPythonEntity for the high level python api.

        Parents are either unresolved references or CaosDB RecordTypes.

        Properties are stored directly as attributes for the object.
        Property metadata is maintained in a dctionary _properties_metadata that should
        never be accessed directly, but only using the get_property_metadata function.
        If property values are references to other objects, they will be stored as
        CaosDBPythonUnresolvedReference objects that can be resolved later into
        CaosDBPythonRecords.
        """

        # Parents are either unresolved references or CaosDB RecordTypes
        self._parents: List[Union[
            CaosDBPythonUnresolvedParent, CaosDBPythonRecordType]] = []
        # self._id: int = CaosDBPythonEntity._get_new_id()
        self._id: Optional[int] = None
        self._name: Optional[str] = None
        self._description: Optional[str] = None
        self._version: Optional[str] = None

        self._file: Optional[str] = None
        self._path: Optional[str] = None

        # name: name of property, value: property metadata
        self._properties_metadata: Dict[CaosDBPropertyMetaData] = dict()

        # Store all current attributes as forbidden attributes
        # which must not be changed by the set_property function.
        self._forbidden = dir(self) + ["_forbidden"]

    def use_parameter(self, name, value):
        self.__setattr__(name, value)
        return value

    @property
    def id(self):
        """
        Getter for the id.
        """
        return self._id

    @id.setter
    def id(self, val: int):
        self._id = val

    @property
    def name(self):
        """
        Getter for the name.
        """
        return self._name

    @name.setter
    def name(self, val: str):
        self._name = val

    @property
    def file(self):
        """
        Getter for the file.
        """
        if type(self) != CaosDBPythonFile:
            raise RuntimeError("Please don't use the file attribute for entities"
                               " that are no files.")
        return self._file

    @file.setter
    def file(self, val: str):
        if val is not None and type(self) != CaosDBPythonFile:
            raise RuntimeError("Please don't use the file attribute for entities"
                               " that are no files.")
        self._file = val

    @property
    def path(self):
        """
        Getter for the path.
        """
        if type(self) != CaosDBPythonFile:
            raise RuntimeError("Please don't use the path attribute for entities"
                               " that are no files.")
        return self._path

    @path.setter
    def path(self, val: str):
        if val is not None and type(self) != CaosDBPythonFile:
            raise RuntimeError("Please don't use the path attribute for entities"
                               " that are no files.")
        self._path = val

    @property
    def description(self):
        """
        Getter for the description.
        """
        return self._description

    @description.setter
    def description(self, val: str):
        self._description = val

    @property
    def version(self):
        """
        Getter for the version.
        """
        return self._version

    @version.setter
    def version(self, val: str):
        self._version = val

    def _set_property_from_entity(self, ent: db.Entity, importance: str,
                                  references: Optional[db.Container],
                                  visited: Dict[int, "CaosDBPythonEntity"]):
        """
        Set a new property using an entity from the normal python API.

        ent : db.Entity
              The entity to be set.
        """

        if ent.name is None:
            raise RuntimeError("Setting properties without name is impossible.")

        if ent.name in self.get_properties():
            raise RuntimeError("Multiproperty not implemented yet.")

        val = self._type_converted_value(ent.value, ent.datatype,
                                         references, visited)
        self.set_property(
            ent.name,
            val,
            datatype=ent.datatype)
        metadata = self.get_property_metadata(ent.name)

        for prop_name in fields(metadata):
            k = prop_name.name
            if k == "importance":
                metadata.importance = importance
            else:
                metadata.__setattr__(k, ent.__getattribute__(k))

    def get_property_metadata(self, prop_name: str) -> CaosDBPropertyMetaData:
        """
        Retrieve the property metadata for the property with name prop_name.

        If the property with the given name does not exist or is forbidden, raise an exception.
        Else return the metadata associated with this property.

        If no metadata does exist yet for the given property, a new object will be created
        and returned.

        prop_name: str
                   Name of the property to retrieve metadata for.
        """

        if not self.property_exists(prop_name):
            raise RuntimeError("The property with name {} does not exist.".format(prop_name))

        if prop_name not in self._properties_metadata:
            self._properties_metadata[prop_name] = CaosDBPropertyMetaData()

        return self._properties_metadata[prop_name]

    def property_exists(self, prop_name: str):
        """
        Check whether a property exists already.
        """
        return prop_name not in self._forbidden and prop_name in self.__dict__

    def set_property(self,
                     name: str,
                     value: Any,
                     overwrite: bool = False,
                     datatype: Optional[str] = None):
        """
        Set a property for this entity with a name and a value.

        If this property is already set convert the value into a list and append the value.
        This behavior can be overwritten using the overwrite flag, which will just overwrite
        the existing value.

        name: str
              Name of the property.

        value: Any
               Value of the property.

        overwrite: bool
                   Use this if you definitely only want one property with
                   that name (set to True).
        """

        if name in self._forbidden:
            raise RuntimeError("Entity cannot be converted to a corresponding "
                               "Python representation. Name of property " +
                               name + " is forbidden!")

        already_exists = self.property_exists(name)

        if already_exists and not overwrite:
            # each call to set_property checks first if it already exists
            #        if yes: Turn the attribute into a list and
            #                place all the elements into that list.
            att = self.__getattribute__(name)

            if isinstance(att, list):
                # just append, see below
                pass
            else:
                old_att = self.__getattribute__(name)
                self.__setattr__(name, [old_att])
            att = self.__getattribute__(name)
            att.append(value)
        else:
            self.__setattr__(name, value)

    def __setattr__(self, name: str, val: Any):
        """
        Allow setting generic properties.
        """

        # TODO: implement checking the value to correspond to one of the datatypes
        #       known for conversion.

        super().__setattr__(name, val)

    def _type_converted_list(self,
                             val: List,
                             pr: str,
                             references: Optional[db.Container],
                             visited: Dict[int, "CaosDBPythonEntity"]):
        """
        Convert a list to a python list of the correct type.

        val: List
             The value of a property containing the list.

        pr: str
            The datatype according to the database entry.
        """
        if not is_list_datatype(pr) and not isinstance(val, list):
            raise RuntimeError("Not a list.")

        return [
            self._type_converted_value(i, get_list_datatype(pr), references,
                                       visited) for i in val]

    def _type_converted_value(self,
                              val: Any,
                              pr: str,
                              references: Optional[db.Container],
                              visited: Dict[int, "CaosDBPythonEntity"]):
        """
        Convert val to the correct type which is indicated by the database
        type string in pr.

        References with ids will be turned into CaosDBPythonUnresolvedReference.
        """

        if val is None:
            return None
        elif isinstance(val, db.Entity):
            # this needs to be checked as second case as it is the ONLY
            # case which does not depend on pr
            # TODO: we might need to pass through the reference container
            return convert_to_python_object(val, references, visited)
        elif isinstance(val, list):
            return self._type_converted_list(val, pr, references, visited)
        elif pr is None:
            return val
        elif pr == DOUBLE:
            return float(val)
        elif pr == BOOLEAN:
            return bool(val)
        elif pr == INTEGER:
            return int(val)
        elif pr == TEXT:
            return str(val)
        elif pr == FILE:
            return CaosDBPythonUnresolvedReference(val)
        elif pr == REFERENCE:
            return CaosDBPythonUnresolvedReference(val)
        elif pr == DATETIME:
            return self._parse_datetime(val)
        elif is_list_datatype(pr):
            return self._type_converted_list(val, pr, references, visited)
        else:
            # Generic references to entities:
            return CaosDBPythonUnresolvedReference(val)

    def _parse_datetime(self, val: Union[str, datetime]):
        """
        Convert val into a datetime object.
        """
        if isinstance(val, datetime):
            return val
        return parser.parse(val)

    def get_property(self, name: str):
        """
        Return the value of the property with name name.

        Raise an exception if the property does not exist.
        """
        if not self.property_exists(name):
            raise RuntimeError("Property {} does not exist.".format(name))
        att = self.__getattribute__(name)
        return att

    def attribute_as_list(self, name: str):
        """
        This is a workaround for the problem that lists containing only one
        element are indistinguishable from simple types in this
        representation.

        TODO: still relevant? seems to be only a problem if LIST types are not used.
        """
        att = self.get_property(name)

        if isinstance(att, list):
            return att
        else:
            return [att]

    def add_parent(self, parent: Union[
            CaosDBPythonUnresolvedParent, "CaosDBPythonRecordType", str]):
        """
        Add a parent to this entity. Either using an unresolved parent or
        using a real record type.

        Strings as argument for parent will automatically be converted to an
        unresolved parent. Likewise, integers as argument will be automatically converted
        to unresolved parents with just an id.
        """

        if isinstance(parent, str):
            parent = CaosDBPythonUnresolvedParent(name=parent)

        if isinstance(parent, int):
            parent = CaosDBPythonUnresolvedParent(id=parent)

        if self.has_parent(parent):
            raise RuntimeError("Duplicate parent.")
        self._parents.append(parent)

    def get_parents(self):
        """
        Returns all parents of this entity.

        Use has_parent for checking for existence of parents
        and add_parent for adding parents to this entity.
        """
        return self._parents

    def has_parent(self, parent: Union[
            CaosDBPythonUnresolvedParent, "CaosDBPythonRecordType"]):
        """
        Check whether this parent already exists for this entity.

        Strings as argument for parent will automatically be converted to an
        unresolved parent. Likewise, integers as argument will be automatically converted
        to unresolved parents with just an id.
        """

        if isinstance(parent, str):
            parent = CaosDBPythonUnresolvedParent(name=parent)

        if isinstance(parent, int):
            parent = CaosDBPythonUnresolvedParent(id=parent)

        for p in self._parents:
            if p.id is not None and p.id == parent.id:
                return True
            elif p.name is not None and p.name == parent.name:
                return True
        return False

    def _resolve_caosdb_python_unresolved_reference(self, propval, deep,
                                                    references, visited):
        # This does not make sense for unset ids:
        if propval.id is None:
            raise RuntimeError("Unresolved property reference without an ID.")
        # have we encountered this id before:
        if propval.id in visited:
            # self.__setattr__(prop, visited[propval.id])
            # don't do the lookup in the references container
            return visited[propval.id]

        if references is None:
            ent = db.Entity(id=propval.id).retrieve()
            obj = convert_to_python_object(ent, references)
            visited[propval.id] = obj
            if deep:
                obj.resolve_references(deep, references, visited)
            return obj

        # lookup in container:
        for ent in references:
            # Entities in container without an ID will be skipped:
            if ent.id is not None and ent.id == propval.id:
                # resolve this entity:
                obj = convert_to_python_object(ent, references)
                visited[propval.id] = obj
                # self.__setattr__(prop, visited[propval.id])
                if deep:
                    obj.resolve_references(deep, references, visited)
                return obj
        return propval

    def resolve_references(self, deep: bool, references: db.Container,
                           visited: Optional[Dict[Union[str, int],
                                                  "CaosDBPythonEntity"]] = None):
        """
        Resolve this entity's references. This affects unresolved properties as well
        as unresolved parents.

        deep: bool
              If True recursively resolve references also for all resolved references.

        references: Optional[db.Container]
                    A container with references that might be resolved.
                    If None is passed as the container, this function tries to resolve entities from a running
                    CaosDB instance directly.
        """

        # This parameter is used in the recursion to keep track of already visited
        # entites (in order to detect cycles).
        if visited is None:
            visited = dict()

        for parent in self.get_parents():
            # TODO
            if isinstance(parent, CaosDBPythonUnresolvedParent):
                pass

        for prop in self.get_properties():
            propval = self.__getattribute__(prop)
            # Resolve all previously unresolved attributes that are entities:
            if deep and isinstance(propval, CaosDBPythonEntity):
                propval.resolve_references(deep, references)
            elif isinstance(propval, list):
                resolvedelements = []
                for element in propval:
                    if deep and isinstance(element, CaosDBPythonEntity):
                        element.resolve_references(deep, references)
                        resolvedelements.append(element)
                    if isinstance(element, CaosDBPythonUnresolvedReference):
                        resolvedelements.append(
                            self._resolve_caosdb_python_unresolved_reference(element, deep,
                                                                             references, visited))
                    else:
                        resolvedelements.append(element)
                self.__setattr__(prop, resolvedelements)

            elif isinstance(propval, CaosDBPythonUnresolvedReference):
                val = self._resolve_caosdb_python_unresolved_reference(propval, deep,
                                                                       references, visited)
                self.__setattr__(prop, val)

    def get_properties(self):
        """
        Return the names of all properties.
        """

        return [p for p in self.__dict__
                if p not in self._forbidden]

    @staticmethod
    def deserialize(serialization: dict):
        """
        Deserialize a yaml representation of an entity in high level API form.
        """

        if "role" in serialization:
            entity = high_level_type_for_role(serialization["role"])()
        else:
            entity = CaosDBPythonRecord()

        if "parents" in serialization:
            for parent in serialization["parents"]:
                if "unresolved" in parent:
                    id = None
                    name = None
                    if "id" in parent:
                        id = parent["id"]
                    if "name" in parent:
                        name = parent["name"]
                    entity.add_parent(CaosDBPythonUnresolvedParent(
                        id=id, name=name))
                else:
                    raise NotImplementedError(
                        "Currently, only unresolved parents can be deserialized.")

        for baseprop in ("name", "id", "description", "version"):
            if baseprop in serialization:
                entity.__setattr__(baseprop, serialization[baseprop])

        if type(entity) == CaosDBPythonFile:
            entity.file = serialization["file"]
            entity.path = serialization["path"]

        for p in serialization["properties"]:
            # The property needs to be set first:

            prop = serialization["properties"][p]
            if isinstance(prop, dict):
                if "unresolved" in prop:
                    entity.__setattr__(p, CaosDBPythonUnresolvedReference(
                        id=prop["id"]))
                else:
                    entity.__setattr__(p,
                                       entity.deserialize(prop))
            else:
                entity.__setattr__(p, prop)

            # if there is no metadata in the yaml file just initialize an empty metadata object
            if "metadata" in serialization and p in serialization["metadata"]:
                metadata = serialization["metadata"][p]
                propmeta = entity.get_property_metadata(p)

                for f in fields(propmeta):
                    if f.name in metadata:
                        propmeta.__setattr__(f.name, metadata[f.name])
            else:
                pass
                # raise NotImplementedError()

        return entity

    def serialize(self, without_metadata: bool = False, visited: dict = None):
        """
        Serialize necessary information into a dict.

        without_metadata: bool
                          If True don't set the metadata field in order to increase
                          readability. Not recommended if deserialization is needed.
        """

        if visited is None:
            visited = dict()

        if self in visited:
            return visited[self]

        metadata: Dict[str, Any] = dict()
        properties = dict()
        parents = list()

        # The full information to be returned:
        fulldict = dict()
        visited[self] = fulldict

        # Add CaosDB role:
        fulldict["role"] = standard_type_for_high_level_type(self, True)

        for parent in self._parents:
            if isinstance(parent, CaosDBPythonEntity):
                parents.append(parent.serialize(without_metadata, visited))
            elif isinstance(parent, CaosDBPythonUnresolvedParent):
                parents.append({"name": parent.name, "id": parent.id,
                                "unresolved": True})
            else:
                raise RuntimeError("Incompatible class used as parent.")

        for baseprop in ("name", "id", "description", "version"):
            val = self.__getattribute__(baseprop)
            if val is not None:
                fulldict[baseprop] = val

        if type(self) == CaosDBPythonFile:
            fulldict["file"] = self.file
            fulldict["path"] = self.path

        for p in self.get_properties():
            m = self.get_property_metadata(p)
            metadata[p] = dict()
            for f in fields(m):
                val = m.__getattribute__(f.name)
                if val is not None:
                    metadata[p][f.name] = val

            val = self.get_property(p)
            if isinstance(val, CaosDBPythonUnresolvedReference):
                properties[p] = {"id": val.id, "unresolved": True}
            elif isinstance(val, CaosDBPythonEntity):
                properties[p] = val.serialize(without_metadata, visited)
            elif isinstance(val, list):
                serializedelements = []
                for element in val:
                    if isinstance(element, CaosDBPythonUnresolvedReference):
                        elm = dict()
                        elm["id"] = element.id
                        elm["unresolved"] = True
                        serializedelements.append(elm)
                    elif isinstance(element, CaosDBPythonEntity):
                        serializedelements.append(
                            element.serialize(without_metadata,
                                              visited))
                    else:
                        serializedelements.append(element)
                properties[p] = serializedelements
            else:
                properties[p] = val

        fulldict["properties"] = properties
        fulldict["parents"] = parents

        if not without_metadata:
            fulldict["metadata"] = metadata
        return fulldict

    def __str__(self):
        return yaml.dump(self.serialize(False))

    # This seemed like a good solution, but makes it difficult to
    # compare python objects directly:
    #
    # def __repr__(self):
    #     return yaml.dump(self.serialize(True))


class CaosDBPythonRecord(CaosDBPythonEntity):
    pass


class CaosDBPythonRecordType(CaosDBPythonEntity):
    pass


class CaosDBPythonProperty(CaosDBPythonEntity):
    pass


class CaosDBMultiProperty:
    """
    This implements a multi property using a python list.
    """

    def __init__(self):
        raise NotImplementedError()


class CaosDBPythonFile(CaosDBPythonEntity):
    def download(self, target=None):
        if self.id is None:
            raise RuntimeError("Cannot download file when id is missing.")
        f = db.File(id=self.id).retrieve()
        return f.download(target)


BASE_ATTRIBUTES = (
    "id", "name", "description", "version", "path", "file")


def _single_convert_to_python_object(robj: CaosDBPythonEntity,
                                     entity: db.Entity,
                                     references: Optional[db.Container] = None,
                                     visited: Optional[Dict[int,
                                                            "CaosDBPythonEntity"]] = None):
    """
    Convert a db.Entity from the standard API to a (previously created)
    CaosDBPythonEntity from the high level API.

    This method will not resolve any unresolved references, so reference properties
    as well as parents will become unresolved references in the first place.

    The optional third parameter can be used
    to resolve references that occur in the converted entities and resolve them
    to their correct representations. (Entities that are not found remain as
    CaosDBPythonUnresolvedReferences.)

    Returns the input object robj.
    """

    # This parameter is used in the recursion to keep track of already visited
    # entites (in order to detect cycles).
    if visited is None:
        visited = dict()

    if id(entity) in visited:
        return visited[id(entity)]
    else:
        visited[id(entity)] = robj

    for base_attribute in BASE_ATTRIBUTES:
        val = entity.__getattribute__(base_attribute)
        if val is not None:
            if isinstance(val, db.common.models.Version):
                val = val.id
            robj.__setattr__(base_attribute, val)

    for prop in entity.properties:
        robj._set_property_from_entity(prop, entity.get_importance(prop), references,
                                       visited)

    for parent in entity.parents:
        robj.add_parent(CaosDBPythonUnresolvedParent(id=parent.id,
                                                     name=parent.name))

    return robj


def _convert_property_value(propval):
    if isinstance(propval, CaosDBPythonUnresolvedReference):
        propval = propval.id
    elif isinstance(propval, CaosDBPythonEntity):
        propval = _single_convert_to_entity(
            standard_type_for_high_level_type(propval)(), propval)
    elif isinstance(propval, list):
        propval = [_convert_property_value(element) for element in propval]

    # TODO: test case for list missing

    return propval


def _single_convert_to_entity(entity: db.Entity,
                              robj: CaosDBPythonEntity):
    """
    Convert a CaosDBPythonEntity to an entity in standard pylib format.

    entity: db.Entity
            An empty entity.

    robj: CaosDBPythonEntity
          The CaosDBPythonEntity that is supposed to be converted to the entity.
    """

    for base_attribute in BASE_ATTRIBUTES:
        if base_attribute in ("file", "path") and not isinstance(robj, CaosDBPythonFile):
            continue

        # Skip version:
        if base_attribute == "version":
            continue

        val = robj.__getattribute__(base_attribute)

        if val is not None:
            entity.__setattr__(base_attribute, val)

    for parent in robj.get_parents():
        if isinstance(parent, CaosDBPythonUnresolvedParent):
            entity.add_parent(name=parent.name, id=parent.id)
        elif isinstance(parent, CaosDBPythonRecordType):
            raise NotImplementedError()
        else:
            raise RuntimeError("Incompatible class used as parent.")

    for prop in robj.get_properties():
        propval = robj.__getattribute__(prop)
        metadata = robj.get_property_metadata(prop)

        propval = _convert_property_value(propval)

        entity.add_property(
            name=prop,
            value=propval,
            unit=metadata.unit,
            importance=metadata.importance,
            datatype=metadata.datatype,
            description=metadata.description,
            id=metadata.id)

    return entity


def convert_to_entity(python_object):
    if isinstance(python_object, db.Container):
        # Create a list of objects:

        return [convert_to_entity(i) for i in python_object]
    elif isinstance(python_object, CaosDBPythonRecord):
        return _single_convert_to_entity(db.Record(), python_object)
    elif isinstance(python_object, CaosDBPythonFile):
        return _single_convert_to_entity(db.File(), python_object)
    elif isinstance(python_object, CaosDBPythonRecordType):
        return _single_convert_to_entity(db.RecordType(), python_object)
    elif isinstance(python_object, CaosDBPythonProperty):
        return _single_convert_to_entity(db.Property(), python_object)
    elif isinstance(python_object, CaosDBPythonEntity):
        return _single_convert_to_entity(db.Entity(), python_object)
    else:
        raise ValueError("Cannot convert an object of this type.")


def convert_to_python_object(entity: Union[db.Container, db.Entity],
                             references: Optional[db.Container] = None,
                             visited: Optional[Dict[int,
                                                    "CaosDBPythonEntity"]] = None):
    """
    Convert either a container of CaosDB entities or a single CaosDB entity
    into the high level representation.

    The optional second parameter can be used
    to resolve references that occur in the converted entities and resolve them
    to their correct representations. (Entities that are not found remain as
    CaosDBPythonUnresolvedReferences.)
    """
    if isinstance(entity, db.Container):
        # Create a list of objects:
        return [convert_to_python_object(i, references, visited) for i in entity]

    # TODO: recursion problems?
    return _single_convert_to_python_object(
        high_level_type_for_standard_type(entity)(),
        entity,
        references,
        visited)


def new_high_level_entity(entity: db.RecordType,
                          importance_level: str,
                          name: Optional[str] = None):
    """
    Create an new record in high level format based on a record type in standard format.

    entity: db.RecordType
            The record type to initialize the new record from.

    importance_level: str
                      None, obligatory, recommended or suggested
                      Initialize new properties up to this level.
                      Properties in the record type with no importance will be added
                      regardless of the importance_level.

    name: str
          Name of the new record.
    """

    r = db.Record(name=name)
    r.add_parent(entity)

    impmap = {
        None: 0, "SUGGESTED": 3, "RECOMMENDED": 2, "OBLIGATORY": 1}

    for prop in entity.properties:
        imp = entity.get_importance(prop)
        if imp is not None and impmap[importance_level] < impmap[imp]:
            continue

        r.add_property(prop)

    return convert_to_python_object(r)


def create_record(rtname: str, name: Optional[str] = None, **kwargs):
    """
    Create a new record based on the name of a record type. The new record is returned.

    rtname: str
            The name of the record type.

    name: str
          This is optional. A name for the new record.

    kwargs:
            Additional arguments are used to set attributes of the
            new record.
    """
    obj = new_high_level_entity(
        db.RecordType(name=rtname).retrieve(), "SUGGESTED", name)
    for key, value in kwargs.items():
        obj.__setattr__(key, value)
    return obj


def load_external_record(record_name: str):
    """
    Retrieve a record by name and convert it to the high level API format.
    """
    return convert_to_python_object(db.Record(name=record_name).retrieve())


def create_entity_container(record: CaosDBPythonEntity):
    """
    Convert this record into an entity container in standard format that can be used
    to insert or update entities in a running CaosDB instance.
    """
    ent = convert_to_entity(record)
    lse: List[db.Entity] = [ent]
    create_flat_list([ent], lse)
    return db.Container().extend(lse)


def query(query: str,
          resolve_references: Optional[bool] = True,
          references: Optional[db.Container] = None):
    """

    """
    res = db.execute_query(query)
    objects = convert_to_python_object(res)
    if resolve_references:
        for obj in objects:
            obj.resolve_references(True, references)
    return objects



# -*- coding: utf-8 -*-
#
# ** header v3.0
# This file is a part of the LinkAhead Project.
#
# Copyright (C) 2018 Research Group Biomedical Physics,
# Max-Planck-Institute for Dynamics and Self-Organization Göttingen
# Copyright (C) 2020 Indiscale GmbH <info@indiscale.com>
# Copyright (C) 2020 Florian Spreckelsen <f.spreckelsen@indiscale.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
# ** end header
#
"""The exceptions module defines exceptions for HTTP Errors (4xx and 5xx and
HTTP response codes) and for transaction errors (i.e. missing permissions,
dependencies, non-passing consistency checks etc.).
"""

from warnings import warn

from lxml import etree


class CaosDBException(Exception):
    # TODO remove from inheritance change once deprecation period is over
    def __init__(self, msg):
        warn(("The name CaosDBException is deprecated. Please use LinkAheadException."),
             DeprecationWarning)
        Exception.__init__(self, msg)
        self.msg = msg


class LinkAheadException(CaosDBException):
    """Base class of all LinkAhead exceptions."""

    def __init__(self, msg):
        Exception.__init__(self, msg)
        self.msg = msg


class ConfigurationError(LinkAheadException):
    """ConfigurationError.

    Indicates a misconfiguration.

    Parameters
    ----------
    msg : str
        A descriptin of the misconfiguration. The constructor adds
        a few lines with explainingg where to find the configuration.

    Attributes
    ----------
    msg : str
        A description of the misconfiguration.
    """

    def __init__(self, msg):
        super().__init__(msg + ConfigurationError._INFO)

    _INFO = ("\n\nPlease check your ~/.pylinkahead.ini and your $PWD/"
             ".pylinkahead.ini. Does at least one of them exist and are they correct?")


class ServerConfigurationException(LinkAheadException):
    """The server is configured in a different way than expected.

    This can be for example unexpected flags or settings or missing extensions.
    """


class HTTPClientError(LinkAheadException):
    """HTTPClientError represents 4xx HTTP client errors."""

    def __init__(self, msg, status, body):
        self.status = status
        self.body = body
        LinkAheadException.__init__(self, msg)


class HTTPServerError(LinkAheadException):
    """HTTPServerError represents 5xx HTTP server errors."""

    def __init__(self, body):
        xml = etree.fromstring(body)
        error = xml.xpath('/Response/Error')[0]
        msg = error.get("description")

        if error.text is not None:
            msg = msg + "\n\n" + error.text
        LinkAheadException.__init__(self, msg)


class CaosDBConnectionError(LinkAheadException):
    # TODO remove from inheritance change once deprecation period is over
    def __init__(self, msg=None):
        warn(("The name CaosDBConnectionError is deprecated. "
              "Please use LinkAheadConnectionError."),
             DeprecationWarning)
        LinkAheadException.__init__(self, msg)


class LinkAheadConnectionError(CaosDBConnectionError):
    """Connection is not configured or the network is down."""

    def __init__(self, msg=None):
        LinkAheadException.__init__(self, msg)


class HTTPURITooLongError(HTTPClientError):
    """The URI of the last request was too long."""

    def __init__(self, msg=None):
        HTTPClientError.__init__(self, msg=msg, status=414, body=None)


class LoginFailedError(LinkAheadException):
    """Login failed.

    Probably, your username/password pair is wrong.
    """

    def __init__(self, msg=None):
        LinkAheadException.__init__(self, msg=msg)


class HTTPForbiddenError(HTTPClientError):
    """You're lacking the required permissions. Corresponds to HTTP status
    403.

    """

    def __init__(self, msg=None):
        HTTPClientError.__init__(self, msg=msg, status=403, body=None)


class HTTPResourceNotFoundError(HTTPClientError):
    """The requested resource doesn't exist; corresponds to HTTP status
    404.

    """

    def __init__(self, msg=None):
        HTTPClientError.__init__(self, msg=msg, status=404, body=None)


class MismatchingEntitiesError(LinkAheadException):
    """Mismatching entities were found during container sync."""


# ######################### Bad query errors ###########################

class BadQueryError(LinkAheadException):
    """Base class for query errors that are not transaction errors."""


class PagingConsistencyError(BadQueryError):
    """The database state changed between two consecutive paged requests of the
    same query."""


class QueryNotUniqueError(BadQueryError):
    """A unique query or retrieve found more than one entity."""


class EmptyUniqueQueryError(BadQueryError):
    """A unique query or retrieve dound no result."""


# ######################### Transaction errors #########################


class TransactionError(LinkAheadException):
    """An error of this type is raised whenever any transaction fails with
    one or more entities between client and LinkAhead server. More
    detailed errors are collected as direct and indirect children in
    the 'errors' list (direct children) and the 'all_errors' set (set
    of all direct and indirect children).

    """

    def __init__(self, error=None,
                 msg="An error occured during the transaction.",
                 container=None):
        LinkAheadException.__init__(self, msg=msg)
        self.errors = []
        self.all_errors = set()
        self.entities = []
        self.all_entities = set()
        self.container = container
        # special case of faulty container
        if container is not None and container.get_errors() is not None:
            self.code = container.get_errors()[0].code
        else:
            self.code = None
        if error is not None:
            self.add_error(error)

    def has_error(self, error_t, direct_children_only=False):
        """Check whether this transaction error contains an error of type
        error_t. If direct_children_only is True, only direct children
        are checked.

        Parameters
        ----------
        error_t : EntityError
            error type to be checked
        direct_children_only: bool, optional
            If True, only direct children, i.e., all errors in
            self.errors are checked. Else all direct and indirect
            children, i.e., all errors in self.all_errors are
            used. Default is false.

        Returns
        -------
        has_error : bool
            True if at least one of the children is of type error_t,
            False otherwise.

        """

        test_set = self.errors if direct_children_only else self.all_errors
        return any([isinstance(err, error_t) for err in test_set])

    def add_error(self, error):
        """Add an error as a direct child to this TransactionError.

        @param error: An EntityError or a list of EntityErrors.

        @raise TypeError: If and only if the 'error' parameter is not an
        instance of EntityError.

        @return: self.
        """

        if hasattr(error, "__iter__"):
            for err in error:
                self.add_error(err)

            return self
        elif isinstance(error, EntityError):
            self.errors.append(error)
            self.entities.append(error.entity)

            self.all_errors.add(error)
            self.all_errors.update(error.all_errors)
            self.all_entities.add(error.entity)
            self.all_entities.update(error.all_entities)

            return self
        else:
            raise TypeError(
                "Argument is to be an EntityError or a list of EntityErrors.")

    def _repr_reasons(self, indent):
        if self.errors is not None and len(self.errors) > 0:
            ret = "\n" + indent + "    +--| REASONS |--"

            for err in self.errors:
                ret += '\n' + indent + '    |  -> ' + \
                    err.__str__(indent=indent + '    |')
            ret += "\n" + indent + "    +----------------"

            return ret
        else:
            return ''

    def _repr_head(self, indent):
        return indent + str(type(self).__name__) + (
            (': ' + self.msg)
            if hasattr(self, 'msg') and self.msg is not None
            else ''
        )

    def __str__(self, indent=''):
        ret = self._repr_head(indent=indent)
        ret += self._repr_reasons(indent=indent)

        return ret

    def __repr__(self):
        return self.__str__()


class EntityError(TransactionError):
    """This is the most basic entity error. It is constructed using an
    entity that caused the error and the error message attached by the
    server.

    """

    def __init__(self, error=None, entity=None):
        TransactionError.__init__(self)
        self.error = error
        if hasattr(error, "code"):
            self.code = error.code
        else:
            self.code = None
        self.entity = entity

        if error is not None and hasattr(error, "encode"):
            self.msg = error
        elif error is not None and hasattr(error, 'description'):
            self.msg = error.description
        elif error is None:
            self.msg = None
        else:
            self.msg = str(error)

    @property
    def description(self):
        """The description of the error."""
        return self.error.description if self.error is not None else None

    def _repr_head(self, indent):
        if hasattr(self, 'entity') and self.entity is not None:
            return (str(type(self.entity).__name__).upper() + " (id: " +
                    str(self.entity.id) + ((", name: " + "'" + str(self.entity.name) + "'") if
                                           self.entity.name is not None else '') + ") CAUSED " +
                    TransactionError._repr_head(self, indent))
        else:
            return TransactionError._repr_head(self, indent)


class UniqueNamesError(EntityError):
    """A name was supposed to be unique but was not."""


class UnqualifiedParentsError(EntityError):
    """This entity has unqualified parents (see 'errors' attribute for a
    list of errors of the parent entities or 'entities' attribute for
    a list of parent entities with errors).

    """


class UnqualifiedPropertiesError(EntityError):
    """This entity has unqualified properties (see 'errors' attribute for
    a list of errors of the properties or 'entities' attribute for a
    list of properties with errors).

    """


class EntityHasNoAclError(EntityError):
    """This entity has no ACL (yet)."""


class EntityDoesNotExistError(EntityError):
    """This entity does not exist."""


class EntityHasNoDatatypeError(EntityError):
    """This has to have a data type."""


class ConsistencyError(EntityError):
    """The transaction violates database consistency."""


class AuthorizationError(EntityError):
    """You are not allowed to do what ever you tried to do.

    Maybe you need more privileges or a user account.
    """


class AmbiguousEntityError(EntityError):
    """A retrieval of the entity was not possible because there is more
    than one possible candidate.
    """



# -*- coding: utf-8 -*-
#
# ** header v3.0
# This file is a part of the LinkAhead Project.
#
# Copyright (C) 2018 Research Group Biomedical Physics,
# Max-Planck-Institute for Dynamics and Self-Organization Göttingen
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
# ** end header
#
"""input.

A CredentialsProvider which reads the password from the input line.
"""
from __future__ import annotations
from .interface import CredentialsProvider, CredentialsAuthenticator
from typing import Optional
import getpass


def get_authentication_provider() -> CredentialsAuthenticator:
    """get_authentication_provider.

    Return an authenticator which uses the input for username/password credentials.

    Returns
    -------
    CredentialsAuthenticator
        with an InputCredentialsProvider as back-end.
    """
    return CredentialsAuthenticator(InputCredentialsProvider())


class InputCredentialsProvider(CredentialsProvider):
    """InputCredentialsProvider.

    A class for obtaining the password directly from the user.

    Methods
    -------
    configure

    Attributes
    ----------
    password
    username
    """

    def __init__(self):
        super(InputCredentialsProvider, self).__init__()
        self._password: Optional[str] = None
        self._username: Optional[str] = None

    def configure(self, **config):
        """configure.

        Parameters
        ----------
        **config
            Keyword arguments containing at least keywords "username" and "password".

        Returns
        -------
        None
        """
        if config.get("username"):
            self._username = config["username"]
        else:
            self._username = input("Please enter the user name: ")

        url = config["url"]
        self._password = getpass.getpass(
            "Please enter the password for `{}` at `{}`: ".format(
                self._username, url))

    @property
    def password(self):
        return self._password

    @property
    def username(self):
        return self._username



# -*- coding: utf-8 -*-
#
# ** header v3.0
# This file is a part of the LinkAhead Project.
#
# Copyright (C) 2018 Research Group Biomedical Physics,
# Max-Planck-Institute for Dynamics and Self-Organization Göttingen
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
# ** end header
#
from __future__ import annotations
import os
import warnings

import yaml

try:
    optional_jsonschema_validate: Optional[Callable] = None
    from jsonschema import validate as optional_jsonschema_validate
except ImportError:
    pass

from configparser import ConfigParser
from os import environ, getcwd
from os.path import expanduser, isfile, join

from typing import Union, Callable, Optional

_pycaosdbconf = ConfigParser(allow_no_value=False)


def _reset_config():
    global _pycaosdbconf
    _pycaosdbconf = ConfigParser(allow_no_value=False)


def configure(inifile: str) -> list[str]:
    """read config from file.

    Return a list of files which have successfully been parsed.
    """
    global _pycaosdbconf
    if ("_pycaosdbconf" not in globals() or _pycaosdbconf is None):
        _reset_config()
    read_config = _pycaosdbconf.read(inifile)
    validate_yaml_schema(config_to_yaml(_pycaosdbconf))

    if "HTTPS_PROXY" in environ:
        _pycaosdbconf["Connection"]["https_proxy"] = environ["HTTPS_PROXY"]
    if "HTTP_PROXY" in environ:
        _pycaosdbconf["Connection"]["http_proxy"] = environ["HTTP_PROXY"]
    return read_config


def get_config() -> ConfigParser:
    global _pycaosdbconf
    if ("_pycaosdbconf" not in globals() or _pycaosdbconf is None):
        _reset_config()
    return _pycaosdbconf


def config_to_yaml(config: ConfigParser) -> dict[str, dict[str, Union[int, str, bool]]]:
    valobj: dict[str, dict[str, Union[int, str, bool]]] = {}
    for s in config.sections():
        valobj[s] = {}
        for key, value in config[s].items():
            # TODO: Can the type be inferred from the config object?
            if key in ["timeout", "debug"]:
                valobj[s][key] = int(value)
            elif key in ["ssl_insecure"]:
                valobj[s][key] = bool(value)
            else:
                valobj[s][key] = value

    return valobj


def validate_yaml_schema(valobj: dict[str, dict[str, Union[int, str, bool]]]):
    if optional_jsonschema_validate:
        with open(os.path.join(os.path.dirname(__file__), "schema-pycaosdb-ini.yml")) as f:
            schema = yaml.load(f, Loader=yaml.SafeLoader)
        optional_jsonschema_validate(instance=valobj, schema=schema["schema-pycaosdb-ini"])
    else:
        warnings.warn("""
            Warning: The validation could not be performed because `jsonschema` is not installed.
        """)


def _read_config_files() -> list[str]:
    """Read config files from different paths.

    Read the config from either ``$PYLINKAHEADINI`` or home directory (``~/.pylinkahead.ini``), and
    additionally adds config from a config file in the current working directory
    (``pylinkahead.ini``).
    If deprecated names are used (starting with 'pycaosdb'), those used in addition but the files
    with the new naming scheme take precedence.
    All of these files are optional.

    Returns
    -------

    ini files: list
      The successfully parsed ini-files. Order: environment variable or home directory, then cwd.
      Used for testing the function.

    """
    return_var = []
    ini_user = expanduser('~/.pylinkahead.ini')
    ini_cwd = join(getcwd(), "pylinkahead.ini")
    # LinkAhead rename block ##################################################
    ini_user_caosdb = expanduser('~/.pycaosdb.ini')
    ini_cwd_caosdb = join(getcwd(), "pycaosdb.ini")
    if os.path.exists(ini_user_caosdb):
        warnings.warn("\n\nYou have a config file with the old naming scheme (pycaosdb.ini). "
                      f"Please use the new version and rename\n"
                      f"    {ini_user_caosdb}\nto\n    {ini_user}", DeprecationWarning)
    if os.path.exists(ini_cwd_caosdb):
        warnings.warn("\n\nYou have a config file with the old naming scheme (pycaosdb.ini). "
                      f"Please use the new version and rename\n"
                      f"    {ini_cwd_caosdb}\nto\n    {ini_cwd}", DeprecationWarning)
    if "PYCAOSDBINI" in environ:
        warnings.warn("\n\nYou have an environment variable PYCAOSDBINI. "
                      "Please rename it to PYLINKAHEADINI.")
    # End: LinkAhead rename block ##################################################

    if "PYLINKAHEADINI" in environ:
        if not isfile(expanduser(environ["PYLINKAHEADINI"])):
            raise RuntimeError(
                f"No configuration file found at\n{expanduser(environ['PYLINKAHEADINI'])}"
                "\nwhich was given via the environment variable PYLINKAHEADINI"
            )
        return_var.extend(configure(expanduser(environ["PYLINKAHEADINI"])))
    else:
        if isfile(ini_user_caosdb):
            return_var.extend(configure(ini_user_caosdb))
        if isfile(ini_user):
            return_var.extend(configure(ini_user))
    if isfile(ini_cwd):
        return_var.extend(configure(ini_cwd))
    if isfile(ini_cwd_caosdb):
        return_var.extend(configure(ini_cwd_caosdb))
    return return_var



# -*- coding: utf-8 -*-
#
# This file is a part of the LinkAhead Project.
#
# Copyright (C) 2018 Research Group Biomedical Physics,
# Max-Planck-Institute for Dynamics and Self-Organization Göttingen
# Copyright (C) 2024 Indiscale GmbH <info@indiscale.com>
# Copyright (C) 2024 Joscha Schmiedt <joscha@schmiedt.dev>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

"""This module provides the interfaces for authenticating requests to the
LinkAhead server.

Implementing modules must provide a `get_authentication_provider()` method.
"""

from __future__ import annotations
from abc import ABC, abstractmethod
import logging
from ..utils import urlencode
from ..interface import CaosDBServerConnection
from ..utils import parse_auth_token, auth_token_to_cookie
from ...exceptions import LoginFailedError
from typing import TYPE_CHECKING, Optional
if TYPE_CHECKING:
    from ..interface import CaosDBHTTPResponse
    QueryDict = dict[str, Optional[str]]


_LOGGER = logging.getLogger(__name__)


class AbstractAuthenticator(ABC):
    """AbstractAuthenticator.

    Interface for different authentication mechanisms. e.g. username/password
    authentication or SSH key authentication.

    Attributes
    ----------
    logger : Logger
        A logger which should be used for all logging which has to do with
        authentication.
    auth_token : str
        A string representation of a LinkAhead Auth Token.

    Methods
    -------
    login (abstract)
    logout (abstract)
    configure (abstract)
    on_request
    on_response

    """

    def __init__(self):
        self.auth_token = None
        self.logger = _LOGGER

    @abstractmethod
    def login(self):
        """login.

        To be implemented by the child classes.

        Returns
        -------
        None
        """
        pass

    @abstractmethod
    def logout(self):
        """logout.

        To be implemented by the child classes.

        Returns
        -------
        None
        """
        pass

    @abstractmethod
    def configure(self, **config):
        """configure.

        Configure this authenticator.

        Parameters
        ----------
        **config
            Keyword arguments for the configuration.

        Returns
        -------
        None
        """
        pass

    def on_response(self, response: CaosDBHTTPResponse):
        """on_response.

        A call-back with is to be called by the connection after each
        response. This method reads the latest auth cookie from the response.

        Parameters
        ----------
        response : CaosDBHTTPResponse
            The response of the server

        Returns
        -------
        """
        self.auth_token = parse_auth_token(
            response.getheader("Set-Cookie"))

    def on_request(self, method: str, path: str, headers: QueryDict, **kwargs):
        # pylint: disable=unused-argument
        """on_request.

        A call-back which is to be called by the connection before each
        request. This method set the auth cookie for that request.

        Parameters
        ----------
        method : str
            The request method.
        path : str
            The request path.
        headers : dict
            A dictionary with headers which are to be set.
        **kwargs
            Ignored

        Returns
        -------
        """
        if self.auth_token is None:
            self.login()
        if self.auth_token is not None:
            headers['Cookie'] = auth_token_to_cookie(self.auth_token)


class CredentialsAuthenticator(AbstractAuthenticator):
    """CredentialsAuthenticator.

    Subclass of AbstractAuthenticator which provides authentication via
    credentials (username/password). This class always needs a
    credentials_provider which provides valid credentials_provider before each
    login.

    Parameters
    ----------
    credentials_provider : CredentialsProvider
        The source for the username and the password.

    Methods
    -------
    login
    logout
    configure
    """

    def __init__(self, credentials_provider):
        super(CredentialsAuthenticator, self).__init__()
        self._credentials_provider = credentials_provider
        self._connection = None
        self.auth_token = None

    def login(self):
        self._login()

    def logout(self):
        self._logout()

    def _logout(self):
        self.logger.debug("[LOGOUT]")
        if self.auth_token is not None:
            self._connection.request(method="DELETE", path="logout")
        self.auth_token = None

    def _login(self):
        username = self._credentials_provider.username
        password = self._credentials_provider.password
        self.logger.debug("[LOGIN] %s", username)

        # we need a username for this:
        if username is None:
            raise LoginFailedError("No username was given.")
        if password is None:
            raise LoginFailedError("No password was given")

        headers = {}
        headers["Content-Type"] = "application/x-www-form-urlencoded"
        body = urlencode({"username": username, "password": password})
        response = self._connection.request(method="POST",
                                            path="login",
                                            headers=headers, body=body)

        response.read()  # clear socket
        if response.status != 200:
            raise LoginFailedError("LOGIN WAS NOT SUCCESSFUL")
        self.on_response(response)
        return response

    def configure(self, **config):
        self._credentials_provider.configure(**config)
        if "connection" in config:
            self._connection = config["connection"]
            if not isinstance(self._connection, CaosDBServerConnection):
                raise Exception("""Bad configuration of the LinkAhead connection.
                                The `connection` must be an instance of
                                `LinkAheadConnection`.""")


class CredentialsProvider(ABC):
    """CredentialsProvider.

    An abstract class for username/password authentication.

    Attributes
    ----------
    password (abstract)
    username (abstract)
    logger : Logger
        A logger which should be used for all logging which has to do with the
        provision of credentials. This is usually just the "authentication"
        logger.

    Methods
    -------
    configure (abstract)
    """

    def __init__(self):
        self.logger = _LOGGER

    @abstractmethod
    def configure(self, **config):
        """configure.

        Configure the credentials provider with a dict.

        Parameters
        ----------
        **config
            Keyword arguments. The relevant arguments depend on the
            implementing subclass of this class.
        Returns
        -------
        None
        """

    @property
    @abstractmethod
    def password(self):
        """password."""

    @property
    @abstractmethod
    def username(self):
        """username."""



# -*- coding: utf-8 -*-
#
# ** header v3.0
# This file is a part of the LinkAhead Project.
#
# Copyright (C) 2018 Research Group Biomedical Physics,
# Max-Planck-Institute for Dynamics and Self-Organization Göttingen
# Copyright (C) 2024 Indiscale GmbH <info@indiscale.com>
# Copyright (C) 2024 Joscha Schmiedt <joscha@schmiedt.dev>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
# ** end header
#
"""Utility functions for the connection module."""
from __future__ import annotations

import re
from urllib.parse import (urlencode as _urlencode, quote as _quote,
                          urlparse, urlunparse, unquote)

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from typing import Optional


def urlencode(query: dict[str, Optional[str]]) -> str:
    """Convert a dict of into a url-encoded (unicode) string.

    This is basically a python2/python3 compatibility wrapper for the respective
    functions in the urllib module with some slight modifications.

    1) A `None` is translated to an empty string.
    >>> urlencode({'key': None})
    'key='

    2) Unicode strings are allowed for python2.
    >>> urlencode({'kèy': 'välüe'})
    'k%C3%A8y=v%C3%A4l%C3%BCe'

    3) All other parameters which can be passed to the respective functions are
    not implemented here and the default parameters will be used.

.. code::

    >>> urlencode({'key': ['val1', 'val2']}, doseq=True)
    Traceback (most recent call last):
        ...
    TypeError: urlencode() got an unexpected keyword argument 'doseq'

Otherwise, this functions works exactly as its counterparts in the urllib
modules when they are called with only the query parameter.

    Parameters
    ----------
    query : dict
        A dict of str or unicode keys with None, unicode or str values.

    Returns
    -------
    str
        A series of `key=value` pairs separated by `&`.

    Raises
    ------
    AttributeError
        If a key or a value does not have a 'encode' method.
    """
    return str(
        _urlencode({
            key.encode("utf-8"): (val.encode("utf-8")
                                  if val is not None else '')
            for key, val in query.items()
        }))


def make_uri_path(segments: Optional[list[str]] = None,
                  query: Optional[dict[str, Optional[str]]] = None) -> str:
    """Url-encode all segments, concat them with slashes and append the query.

    Examples
    --------
    >>> make_uri_path(['a','b'])
    'a/b'
    >>> make_uri_path(['a','ö'])
    'a/%C3%B6'
    >>> make_uri_path(['a','b'], {'key': 'val'})
    'a/b?key=val'

    Parameters
    ----------
    segments : list of str
        The segments of the path
    query: dict
        A dict of str keys with None or str values.

    Returns
    -------
    str
        A relative uri path (no host information, possibly no root path).
    """
    path_no_query = ("/".join([quote(segment) for segment in segments])
                     if segments else "")
    if query is None:
        return str(path_no_query)

    return str("?".join([
        path_no_query, "&".join([
            quote(key) + "=" +
            (quote(query[key]) if query[key] is not None else "")  # type: ignore
            for key in query
        ])
    ]))


def quote(string: str) -> str:
    enc = string.encode('utf-8')
    return _quote(enc).replace('/', '%2F')


def parse_url(url: str):
    fullurl = urlparse(url=url)
    # make sure the path ends with a slash
    if not fullurl.path.endswith("/"):
        parse_result = list(fullurl)
        parse_result[2] += "/"
        fullurl = urlparse(urlunparse(parse_result))
    return fullurl


_PATTERN = re.compile(r"^SessionToken=([^;]*);.*$")


def parse_auth_token(cookie: Optional[str]) -> Optional[str]:
    """parse_auth_token.

    Parse an auth token from a cookie.

    Parameters
    ----------
    cookie : str
        A cookie with an urlencoded authtoken.

    Returns
    -------
    str
        An auth token string.
    """
    auth_token = None
    if cookie is not None and _PATTERN.match(cookie):
        auth_token = unquote(_PATTERN.split(cookie)[1])
    return auth_token


def auth_token_to_cookie(auth_token: str) -> str:
    """auth_token_to_cookie.

    Urlencode an auth token string and format it as a cookie.

    Parameters
    ----------
    auth_token : str
        The plain auth token string.

    Raises
    ------
    TypeError
        If the auth_token was None

    Returns
    -------
    str
        A cookie
    """
    if auth_token is None:
        raise TypeError("Parameter `auth_token` was None.")
    return "SessionToken=" + quote(auth_token) + ";"



# -*- encoding: utf-8 -*-
#
# ** header v3.0
# This file is a part of the LinkAhead Project.
#
# Copyright (C) 2018 Research Group Biomedical Physics,
# Max-Planck-Institute for Dynamics and Self-Organization Göttingen
# Copyright (C) 2024 Indiscale GmbH <info@indiscale.com>
# Copyright (C) 2024 Joscha Schmiedt <joscha@schmiedt.dev>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
# ** end header
#
# Original work Copyright (c) 2011 Chris AtLee
# Modified work Copyright (c) 2017 Biomedical Physics, MPI for Dynamics and Self-Organization
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""multipart/form-data encoding module.

This module provides functions that faciliate encoding name/value pairs
as multipart/form-data suitable for a HTTP POST or PUT request.

multipart/form-data is the standard way to upload files over HTTP
"""
from __future__ import annotations

__all__ = [
    'gen_boundary', 'encode_and_quote', 'MultipartParam', 'encode_string',
    'encode_file_header', 'get_body_size', 'get_headers', 'multipart_encode',
    'ReadableMultiparts',
]
from urllib.parse import quote_plus
from io import UnsupportedOperation
import uuid
import re
import os
import mimetypes
from email.header import Header
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Optional


def gen_boundary():
    """Returns a random string to use as the boundary for a message."""
    return uuid.uuid4().hex


def encode_and_quote(data: Optional[str]) -> Optional[str]:
    """If ``data`` is unicode, return urllib.quote_plus(data.encode("utf-8"))
    otherwise return urllib.quote_plus(data)"""
    if data is None:
        return None

    return quote_plus(data)


class MultipartParam(object):
    """Represents a single parameter in a multipart/form-data request.

    ``name`` is the name of this parameter.

    If ``value`` is set, it must be a string or unicode object to use as the
    data for this parameter.

    If ``filename`` is set, it is what to say that this parameter's filename
    is.  Note that this does not have to be the actual filename any local file.

    If ``filetype`` is set, it is used as the Content-Type for this parameter.
    If unset it defaults to "text/plain; charset=utf8"

    If ``filesize`` is set, it specifies the length of the file ``fileobj``

    If ``fileobj`` is set, it must be a file-like object that supports
    .read().

    Both ``value`` and ``fileobj`` must not be set, doing so will
    raise a ValueError assertion.

    If ``fileobj`` is set, and ``filesize`` is not specified, then
    the file's size will be determined first by stat'ing ``fileobj``'s
    file descriptor, and if that fails, by seeking to the end of the file,
    recording the current position as the size, and then by seeking back to the
    beginning of the file.

    ``callback`` is a callable which will be called from iter_encode with (self,
    current_transferred, total), representing the current parameter, current amount
    transferred, and the total size.
    """

    def __init__(self,
                 name: str,
                 value=None,
                 filename=None,
                 filetype=None,
                 filesize=None,
                 fileobj=None,
                 callback=None):
        self.name = Header(name).encode()
        self.value = value
        if filename is None:
            self.filename = None
        else:
            bfilename = filename.encode("ascii", "xmlcharrefreplace")
            self.filename = bfilename.decode("UTF-8").replace('"', '\\"')

        self.filetype = filetype

        self.filesize = filesize
        self.fileobj = fileobj
        self.callback = callback

        if self.value is not None and self.fileobj is not None:
            raise ValueError("Only one of value or fileobj may be specified")

        if fileobj is not None and filesize is None:
            # Try and determine the file size
            try:
                self.filesize = os.fstat(fileobj.fileno()).st_size
            except (OSError, AttributeError, UnsupportedOperation):
                try:
                    fileobj.seek(0, 2)
                    self.filesize = fileobj.tell()
                    fileobj.seek(0)
                except BaseException:
                    raise ValueError("Could not determine filesize")

    def reset(self):
        """Reset the file object's read pointer."""
        if self.fileobj is not None:
            self.fileobj.seek(0)
        elif self.value is None:
            raise ValueError("Don't know how to reset this parameter")

    @classmethod
    def from_file(cls, paramname, filename):
        """Returns a new MultipartParam object constructed from the local file
        at ``filename``.

        ``filesize`` is determined by os.path.getsize(``filename``)

        ``filetype`` is determined by mimetypes.guess_type(``filename``)[0]

        ``filename`` is set to os.path.basename(``filename``)
        """

        return cls(
            paramname,
            filename=os.path.basename(filename),
            filetype=mimetypes.guess_type(filename)[0],
            filesize=os.path.getsize(filename),
            fileobj=open(filename, "rb"))

    @classmethod
    def from_params(cls, params):
        """Returns a list of MultipartParam objects from a sequence of name,
        value pairs, MultipartParam instances, or from a mapping of names to
        values.

        The values may be strings or file objects, or MultipartParam
        objects. MultipartParam object names must match the given names
        in the name,value pairs or mapping, if applicable.
        """
        if hasattr(params, 'items'):
            params = params.items()

        retval = []
        for item in params:
            if isinstance(item, cls):
                retval.append(item)
                continue
            name, value = item
            if isinstance(value, cls):
                assert value.name == name
                retval.append(value)
                continue
            if hasattr(value, 'read'):
                # Looks like a file object
                filename = getattr(value, 'name', None)
                if filename is not None:
                    filetype = mimetypes.guess_type(filename)[0]
                else:
                    filetype = None

                retval.append(
                    cls(name=name,
                        filename=filename,
                        filetype=filetype,
                        fileobj=value))
            else:
                retval.append(cls(name, value))
        return retval

    def encode_hdr(self, boundary):
        """Returns the header of the encoding of this parameter."""
        boundary = encode_and_quote(boundary)

        headers = ["--%s" % boundary]

        if self.filename:
            disposition = 'form-data; name="%s"; filename="%s"' % (
                self.name, self.filename)
        else:
            disposition = 'form-data; name="%s"' % self.name

        headers.append("Content-Disposition: %s" % disposition)

        if self.filetype:
            filetype = self.filetype
        else:
            filetype = "text/plain; charset=utf-8"

        headers.append("Content-Type: %s" % filetype)

        headers.append("")
        headers.append("")

        return "\r\n".join(headers)

    def encode(self, boundary):
        """Returns the string encoding of this parameter."""
        if self.value is None:
            value = self.fileobj.read()
        else:
            value = self.value

        if re.search("^--%s$" % re.escape(boundary), value, re.M):
            raise ValueError("boundary found in encoded string")

        return "%s%s\r\n" % (self.encode_hdr(boundary), value)

    def iter_encode(self, boundary, blocksize=4096):
        """Yields the encoding of this parameter If self.fileobj is set, then
        blocks of ``blocksize`` bytes are read and yielded."""
        total = self.get_size(boundary)
        current_transferred = 0
        if self.value is not None:
            block = self.encode(boundary)
            current_transferred += len(block)
            yield block
            if self.callback:
                self.callback(self, current_transferred, total)
        else:
            block = self.encode_hdr(boundary)
            current_transferred += len(block)
            yield block
            if self.callback:
                self.callback(self, current_transferred, total)
            last_block = b""
            encoded_boundary = "--%s" % encode_and_quote(boundary)
            boundary_exp = re.compile("^%s$" % re.escape(encoded_boundary),
                                      re.M)
            while True:
                block = self.fileobj.read(blocksize)
                if not block:
                    current_transferred += 2
                    yield "\r\n"
                    if self.callback:
                        self.callback(self, current_transferred, total)
                    break
                last_block += block
                if boundary_exp.search(last_block.decode("ascii", "ignore")):
                    raise ValueError("boundary found in file data")
                last_block = last_block[-len(encoded_boundary) - 2:]
                current_transferred += len(block)
                yield block
                if self.callback:
                    self.callback(self, current_transferred, total)

    def get_size(self, boundary):
        """Returns the size in bytes that this param will be when encoded with
        the given boundary."""
        if self.filesize is not None:
            valuesize = self.filesize
        else:
            valuesize = len(self.value)

        return len(self.encode_hdr(boundary)) + 2 + valuesize


def encode_string(boundary, name, value):
    """Returns ``name`` and ``value`` encoded as a multipart/form-data
    variable.

    ``boundary`` is the boundary string used throughout a single request
    to separate variables.
    """

    return MultipartParam(name, value).encode(boundary)


def encode_file_header(boundary,
                       paramname,
                       filesize,
                       filename=None,
                       filetype=None):
    """Returns the leading data for a multipart/form-data field that contains
    file data.

    ``boundary`` is the boundary string used throughout a single request to
    separate variables.

    ``paramname`` is the name of the variable in this request.

    ``filesize`` is the size of the file data.

    ``filename`` if specified is the filename to give to this field.  This
    field is only useful to the server for determining the original filename.

    ``filetype`` if specified is the MIME type of this file.

    The actual file data should be sent after this header has been sent.
    """

    return MultipartParam(
        paramname, filesize=filesize, filename=filename,
        filetype=filetype).encode_hdr(boundary)


def get_body_size(params, boundary):
    """Returns the number of bytes that the multipart/form-data encoding of
    ``params`` will be."""
    size = sum(
        p.get_size(boundary) for p in MultipartParam.from_params(params))
    return size + len(boundary) + 6


def get_headers(params, boundary):
    """Returns a dictionary with Content-Type and Content-Length headers for
    the multipart/form-data encoding of ``params``."""
    headers = {}
    boundary = quote_plus(boundary)
    headers['Content-Type'] = "multipart/form-data; boundary=%s" % boundary
    headers['Content-Length'] = str(get_body_size(params, boundary))
    return headers


class MultipartYielder(object):
    """An iterator that yields the parameters of a multipart/formdata http
    body."""

    def __init__(self, params, boundary, callback):
        self.params = params
        self.boundary = boundary
        self.callback = callback

        self.i = 0
        self.current_part = None
        self.param_iter = None
        self.current_transferred = 0
        self.total = get_body_size(params, boundary)

    def __iter__(self):
        return self

    # since python 3
    def __next__(self):
        return self.next()

    def next(self):
        """generator function to yield multipart/form-data representation of
        parameters."""
        if self.param_iter is not None:
            try:
                block = next(self.param_iter)
                self.current_transferred += len(block)
                if self.callback:
                    self.callback(self.current_part,
                                  self.current_transferred, self.total)
                return block
            except StopIteration:
                self.current_part = None
                self.param_iter = None

        if self.i is None:
            raise StopIteration
        elif self.i >= len(self.params):
            self.param_iter = None
            self.current_part = None
            self.i = None
            block = "--%s--\r\n" % self.boundary
            self.current_transferred += len(block)
            if self.callback:
                self.callback(self.current_part,
                              self.current_transferred, self.total)
            return block

        self.current_part = self.params[self.i]
        self.param_iter = self.current_part.iter_encode(self.boundary)
        self.i += 1
        return next(self)

    def reset(self):
        """Reset the iterator."""
        self.i = 0
        self.current_transferred = 0
        for param in self.params:
            param.reset()


def multipart_encode(params, boundary=None, callback=None):
    """Encode ``params`` as multipart/form-data.

    ``params`` should be a sequence of (name, value) pairs or MultipartParam
    objects, or a mapping of names to values.
    Values are either strings parameter values, or file-like objects to use as
    the parameter value.  The file-like objects must support .read() and either
    .fileno() or both .seek() and .tell().

    If ``boundary`` is set, then it as used as the MIME boundary.  Otherwise
    a randomly generated boundary will be used.  In either case, if the
    boundary string appears in the parameter values a ValueError will be
    raised.

    If ``callback`` is set, it should be a callback which will get called as blocks
    of data are encoded.  It will be called with (param, current_transferred, total),
    indicating the current parameter being encoded, the current amount encoded,
    and the total amount to encode.

    Returns a tuple of `datagen`, `headers`, where `datagen` is a
    generator that will yield blocks of data that make up the encoded
    parameters, and `headers` is a dictionary with the assoicated
    Content-Type and Content-Length headers.

    Examples:

    >>> datagen, headers = multipart_encode( [("key", "value1"), ("key", "value2")] )
    >>> s = "".join(datagen)
    >>> assert "value2" in s and "value1" in s

    >>> p = MultipartParam("key", "value2")
    >>> datagen, headers = multipart_encode( [("key", "value1"), p] )
    >>> s = "".join(datagen)
    >>> assert "value2" in s and "value1" in s

    >>> datagen, headers = multipart_encode( {"key": "value1"} )
    >>> s = "".join(datagen)
    >>> assert "value2" not in s and "value1" in s
    """
    if boundary is None:
        boundary = gen_boundary()
    else:
        boundary = quote_plus(boundary)

    headers = get_headers(params, boundary)
    params = MultipartParam.from_params(params)

    return MultipartYielder(params, boundary, callback), headers


class ReadableMultiparts(object):
    """Wraps instances of the MultipartYielder class as a readable and withable
    object."""

    def __init__(self, multipart_yielder):
        self.multipart_yielder = multipart_yielder
        self.current_block = None
        self.left_over = b''

    def read(self, size=-1):
        result = self.left_over
        while size == -1 or len(result) < size:
            try:
                next_chunk = self.multipart_yielder.next()
                if hasattr(next_chunk, "encode"):
                    next_chunk = next_chunk.encode("utf8")
                result += next_chunk
            except StopIteration:
                break

        if size == -1:
            self.left_over = b''
            return result

        self.left_over = result[size:]
        return result[:size]

    def __enter__(self):
        pass

    def __exit__(self, type, value, traceback):
        self.close()

    def close(self):
        self.multipart_yielder.reset()



# -*- coding: utf-8 -*-
#
# This file is a part of the LinkAhead Project.
#
# Copyright (C) 2018 Research Group Biomedical Physics,
# Max-Planck-Institute for Dynamics and Self-Organization Göttingen
# Copyright (C) 2020 Timm Fitschen <t.fitschen@indiscale.com>
# Copyright (C) 2020-2022 IndiScale GmbH <info@indiscale.com>
# Copyright (C) 2024 Indiscale GmbH <info@indiscale.com>
# Copyright (C) 2024 Joscha Schmiedt <joscha@schmiedt.dev>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
# ** end header
#
"""API-Utils: Some simplified functions for generation of records etc.

"""
from __future__ import annotations
import logging
import warnings
from collections.abc import Iterable
from typing import Any, Union, Optional

from .common.datatype import is_reference
from .common.models import (SPECIAL_ATTRIBUTES, Container, Entity, File,
                            Property, Record, RecordType, execute_query)
from .exceptions import LinkAheadException
from .utils.git_utils import (get_branch_in, get_commit_in, get_diff_in,
                              get_origin_url_in)

logger = logging.getLogger(__name__)


class EntityMergeConflictError(LinkAheadException):
    """An error that is raised in case of an unresolvable conflict when merging
    two entities.
    """


def new_record(record_type: Union[str],
               name: Optional[str] = None,
               description: Optional[str] = None,
               tempid: Optional[int] = None,
               insert: bool = False, **kwargs) -> Record:
    """Function to simplify the creation of Records.

    record_type: The name of the RecordType to use for this record.
    name: Name of the new Record.
    kwargs: Key-value-pairs for the properties of this Record.

    Returns: The newly created Record.

    Of course this functions requires an open database connection!
    """

    rt = RecordType(name=record_type)
    rt.retrieve()

    r = Record(name)
    r.add_parent(rt)

    if tempid is not None:
        r.id = tempid

    if description is not None:
        r.description = description

    # Add all additional properties, treat iterables als multiple
    # additions.

    for k, v in kwargs.items():
        if hasattr(v, "encode") or not isinstance(v, Iterable):
            v = [v]

        for vv in v:
            p = Property(k)
            p.retrieve()
            p.value = vv
            r.add_property(p)

    if insert:
        r.insert()

    return r


def id_query(ids: list[int]) -> Container:
    warnings.warn("Please use 'create_id_query', which only creates"
                  "the string.", DeprecationWarning)

    return execute_query(create_id_query(ids))  # type: ignore


def create_id_query(ids: list[int]) -> str:
    return "FIND ENTITY WITH " + " OR ".join(
        ["ID={}".format(id) for id in ids])


def get_type_of_entity_with(id_: int):
    objs = retrieve_entities_with_ids([id_])

    if len(objs) == 0:
        raise RuntimeError("ID {} not found.".format(id_))

    if len(objs) > 1:
        raise RuntimeError(
            "ID {} is not unique. This is probably a bug in the LinkAhead server." .format(id_))
    obj = objs[0]

    if isinstance(obj, Record):
        return Record
    elif isinstance(obj, RecordType):
        return RecordType
    elif isinstance(obj, Property):
        return Property
    elif isinstance(obj, File):
        return File
    elif isinstance(obj, Entity):
        return Entity


def retrieve_entity_with_id(eid: int):
    return execute_query("FIND ENTITY WITH ID={}".format(eid), unique=True)


def retrieve_entities_with_ids(entities: list) -> Container:
    collection = Container()
    step = 20

    for i in range(len(entities)//step+1):
        collection.extend(
            execute_query(
                create_id_query(entities[i*step:(i+1)*step])))

    return collection


def getOriginUrlIn(folder):
    warnings.warn("""
                  This function is deprecated and will be removed with the next release.
                  Please use the linkahead.utils.git_utils.get_origin_url_in instead.""",
                  DeprecationWarning)
    return get_origin_url_in(folder)


def getDiffIn(folder, save_dir=None):
    warnings.warn("""
                  This function is deprecated and will be removed with the next release.
                  Please use the linkahead.utils.git_utils.get_diff_in instead.""",
                  DeprecationWarning)
    return get_diff_in(folder, save_dir)


def getBranchIn(folder):
    warnings.warn("""
                  This function is deprecated and will be removed with the next release.
                  Please use the linkahead.utils.git_utils.get_branch_in instead.""",
                  DeprecationWarning)
    return get_branch_in(folder)


def getCommitIn(folder):
    warnings.warn("""
                  This function is deprecated and will be removed with the next release.
                  Please use the linkahead.utils.git_utils.get_commit_in instead.""",
                  DeprecationWarning)
    return get_commit_in(folder)


def compare_entities(old_entity: Entity,
                     new_entity: Entity,
                     compare_referenced_records: bool = False
                     ) -> tuple[dict[str, Any], dict[str, Any]]:
    """Compare two entites.

    Return a tuple of dictionaries, the first index belongs to additional information for old
    entity, the second index belongs to additional information for new entity.

    Additional information means in detail:
    - Additional parents (a list under key "parents")
    - Information about properties:
      - Each property lists either an additional property or a property with a changed:
        - datatype
        - importance or
        - value (not implemented yet)

        In case of changed information the value listed under the respective key shows the
        value that is stored in the respective entity.

    If `compare_referenced_records` is `True`, also referenced entities will be
    compared using this function (which is then called with
    `compare_referenced_records = False` to prevent infinite recursion in case
    of circular references).

    Parameters
    ----------
    old_entity, new_entity : Entity
        Entities to be compared
    compare_referenced_records : bool, optional
        Whether to compare referenced records in case of both, `old_entity` and
        `new_entity`, have the same reference properties and both have a Record
        object as value. If set to `False`, only the corresponding Python
        objects are compared which may lead to unexpected behavior when
        identical records are stored in different objects. Default is False.

    """
    olddiff: dict[str, Any] = {"properties": {}, "parents": []}
    newdiff: dict[str, Any] = {"properties": {}, "parents": []}

    if old_entity is new_entity:
        return (olddiff, newdiff)

    if type(old_entity) is not type(new_entity):
        raise ValueError(
            "Comparison of different Entity types is not supported.")

    for attr in SPECIAL_ATTRIBUTES:
        try:
            oldattr = old_entity.__getattribute__(attr)
            old_entity_attr_exists = True
        except BaseException:
            old_entity_attr_exists = False
        try:
            newattr = new_entity.__getattribute__(attr)
            new_entity_attr_exists = True
        except BaseException:
            new_entity_attr_exists = False

        if old_entity_attr_exists and (oldattr == "" or oldattr is None):
            old_entity_attr_exists = False

        if new_entity_attr_exists and (newattr == "" or newattr is None):
            new_entity_attr_exists = False

        if not old_entity_attr_exists and not new_entity_attr_exists:
            continue

        if ((old_entity_attr_exists ^ new_entity_attr_exists)
                or (oldattr != newattr)):

            if old_entity_attr_exists:
                olddiff[attr] = oldattr

            if new_entity_attr_exists:
                newdiff[attr] = newattr

    # properties

    for prop in old_entity.properties:
        matching = [p for p in new_entity.properties if p.name == prop.name]

        if len(matching) == 0:
            olddiff["properties"][prop.name] = {}
        elif len(matching) == 1:
            newdiff["properties"][prop.name] = {}
            olddiff["properties"][prop.name] = {}

            if (old_entity.get_importance(prop.name) !=
                    new_entity.get_importance(prop.name)):
                olddiff["properties"][prop.name]["importance"] = \
                    old_entity.get_importance(prop.name)
                newdiff["properties"][prop.name]["importance"] = \
                    new_entity.get_importance(prop.name)

            if (prop.datatype != matching[0].datatype):
                olddiff["properties"][prop.name]["datatype"] = prop.datatype
                newdiff["properties"][prop.name]["datatype"] = \
                    matching[0].datatype

            if (prop.unit != matching[0].unit):
                olddiff["properties"][prop.name]["unit"] = prop.unit
                newdiff["properties"][prop.name]["unit"] = \
                    matching[0].unit

            if (prop.value != matching[0].value):
                # basic comparison of value objects says they are different
                same_value = False
                if compare_referenced_records:
                    # scalar reference
                    if isinstance(prop.value, Entity) and isinstance(matching[0].value, Entity):
                        # explicitely not recursive to prevent infinite recursion
                        same_value = empty_diff(
                            prop.value, matching[0].value, compare_referenced_records=False)
                    # list of references
                    elif isinstance(prop.value, list) and isinstance(matching[0].value, list):
                        # all elements in both lists actually are entity objects
                        # TODO: check, whether mixed cases can be allowed or should lead to an error
                        if (all([isinstance(x, Entity) for x in prop.value])
                                and all([isinstance(x, Entity) for x in matching[0].value])):
                            # can't be the same if the lengths are different
                            if len(prop.value) == len(matching[0].value):
                                # do a one-by-one comparison:
                                # the values are the same if all diffs are empty
                                same_value = all(
                                    [empty_diff(x, y, False) for x, y
                                     in zip(prop.value, matching[0].value)])

                if not same_value:
                    olddiff["properties"][prop.name]["value"] = prop.value
                    newdiff["properties"][prop.name]["value"] = \
                        matching[0].value

            if (len(newdiff["properties"][prop.name]) == 0
                    and len(olddiff["properties"][prop.name]) == 0):
                newdiff["properties"].pop(prop.name)
                olddiff["properties"].pop(prop.name)

        else:
            raise NotImplementedError(
                "Comparison not implemented for multi-properties.")

    for prop in new_entity.properties:
        if len([0 for p in old_entity.properties if p.name == prop.name]) == 0:
            newdiff["properties"][prop.name] = {}

    # parents

    for parent in old_entity.parents:
        if len([0 for p in new_entity.parents if p.name == parent.name]) == 0:
            olddiff["parents"].append(parent.name)

    for parent in new_entity.parents:
        if len([0 for p in old_entity.parents if p.name == parent.name]) == 0:
            newdiff["parents"].append(parent.name)

    return (olddiff, newdiff)


def empty_diff(old_entity: Entity, new_entity: Entity,
               compare_referenced_records: bool = False) -> bool:
    """Check whether the `compare_entities` found any differences between
    old_entity and new_entity.

    Parameters
    ----------
    old_entity, new_entity : Entity
        Entities to be compared
    compare_referenced_records : bool, optional
        Whether to compare referenced records in case of both, `old_entity` and
        `new_entity`, have the same reference properties and both have a Record
        object as value.

    """
    olddiff, newdiff = compare_entities(
        old_entity, new_entity, compare_referenced_records)
    for diff in [olddiff, newdiff]:
        for key in ["parents", "properties"]:
            if len(diff[key]) > 0:
                # There is a difference somewhere in the diff
                return False
        for key in SPECIAL_ATTRIBUTES:
            if key in diff and diff[key]:
                # There is a difference in at least one special attribute
                return False
    # all elements of the two diffs were empty
    return True


def merge_entities(entity_a: Entity,
                   entity_b: Entity,
                   merge_references_with_empty_diffs=True,
                   force=False,
                   merge_id_with_resolved_entity: bool = False
                   ) -> Entity:
    """Merge entity_b into entity_a such that they have the same parents and properties.

    datatype, unit, value, name and description will only be changed in entity_a
    if they are None for entity_a and set for entity_b. If there is a
    corresponding value for entity_a different from None, an
    EntityMergeConflictError will be raised to inform about an unresolvable merge
    conflict.

    The merge operation is done in place.

    Returns entity_a.

    WARNING: This function is currently experimental and insufficiently tested. Use with care.

    Parameters
    ----------
    entity_a, entity_b : Entity
        The entities to be merged. entity_b will be merged into entity_a in place
    merge_references_with_empty_diffs : bool, optional
        Whether the merge is performed if entity_a and entity_b both reference
        record(s) that may be different Python objects but have empty diffs. If
        set to `False` a merge conflict will be raised in this case
        instead. Default is True.
    force : bool, optional
        If True, in case `entity_a` and `entity_b` have the same properties, the
        values of `entity_a` are replaced by those of `entity_b` in the
        merge. If `False`, an EntityMergeConflictError is raised
        instead. Default is False.
    merge_id_with_resolved_entity : bool, optional
        If true, the values of two reference properties will be considered the
        same if one is an integer id and the other is a db.Entity with this
        id. I.e., a value 123 is identified with a value ``<Record
        id=123/>``. Default is False.

    Returns
    -------
    entity_a : Entity
       The initial entity_a after the in-place merge

    Raises
    ------
    EntityMergeConflictError
        In case of an unresolvable merge conflict.

    """

    logger.warning(
        "This function is currently experimental and insufficiently tested. Use with care.")

    # Compare both entities:
    diff_r1, diff_r2 = compare_entities(
        entity_a, entity_b, compare_referenced_records=merge_references_with_empty_diffs)

    # Go through the comparison and try to apply changes to entity_a:
    for key in diff_r2["parents"]:
        entity_a.add_parent(entity_b.get_parent(key))

    for key in diff_r2["properties"]:
        if key in diff_r1["properties"]:
            if ("importance" in diff_r1["properties"][key] and
                    "importance" in diff_r2["properties"][key]):
                if (diff_r1["properties"][key]["importance"] !=
                        diff_r2["properties"][key]["importance"]):
                    raise NotImplementedError()
            elif ("importance" in diff_r1["properties"][key] or
                  "importance" in diff_r2["properties"][key]):
                raise NotImplementedError()

            for attribute in ("datatype", "unit", "value"):
                if (attribute in diff_r2["properties"][key] and
                        diff_r2["properties"][key][attribute] is not None):
                    if (diff_r1["properties"][key][attribute] is None):
                        setattr(entity_a.get_property(key), attribute,
                                diff_r2["properties"][key][attribute])
                    elif force:
                        setattr(entity_a.get_property(key), attribute,
                                diff_r2["properties"][key][attribute])
                    else:
                        raise_error = True
                        if merge_id_with_resolved_entity is True and attribute == "value":
                            # Do a special check for the case of an id value on the
                            # one hand, and a resolved entity on the other side.
                            prop_a = entity_a.get_property(key)
                            assert prop_a is not None, f"Property {key} not found in entity_a"
                            prop_b = entity_b.get_property(key)
                            assert prop_b is not None, f"Property {key} not found in entity_b"
                            this = prop_a.value
                            that = prop_b.value
                            same = False
                            if isinstance(this, list) and isinstance(that, list):
                                if len(this) == len(that):
                                    same = all([_same_id_as_resolved_entity(a, b)
                                                for a, b in zip(this, that)])
                            else:
                                same = _same_id_as_resolved_entity(this, that)
                            if same is True:
                                setattr(entity_a.get_property(key), attribute,
                                        diff_r2["properties"][key][attribute])
                                raise_error = False
                        if raise_error is True:
                            raise EntityMergeConflictError(
                                f"Entity a ({entity_a.id}, {entity_a.name}) "
                                f"has a Property '{key}' with {attribute}="
                                f"{diff_r2['properties'][key][attribute]}\n"
                                f"Entity b ({entity_b.id}, {entity_b.name}) "
                                f"has a Property '{key}' with {attribute}="
                                f"{diff_r1['properties'][key][attribute]}")
        else:
            # TODO: This is a temporary FIX for
            #       https://gitlab.indiscale.com/caosdb/src/caosdb-pylib/-/issues/105
            prop_b = entity_b.get_property(key)
            assert prop_b is not None, f"Property {key} not found in entity_b"
            entity_a.add_property(id=prop_b.id,
                                  name=prop_b.name,
                                  datatype=prop_b.datatype,
                                  value=prop_b.value,
                                  unit=prop_b.unit,
                                  importance=entity_b.get_importance(key))
            # entity_a.add_property(
            #     entity_b.get_property(key),
            #     importance=entity_b.get_importance(key))

    for special_attribute in ("name", "description"):
        sa_a = getattr(entity_a, special_attribute)
        sa_b = getattr(entity_b, special_attribute)
        if sa_a != sa_b:
            if sa_a is None:
                setattr(entity_a, special_attribute, sa_b)
            elif force:
                # force overwrite
                setattr(entity_a, special_attribute, sa_b)
            else:
                raise EntityMergeConflictError(
                    f"Conflict in special attribute {special_attribute}:\n"
                    f"A: {sa_a}\nB: {sa_b}")
    return entity_a


def describe_diff(olddiff, newdiff, name=None, as_update=True):
    description = ""

    for attr in list(set(list(olddiff.keys()) + list(newdiff.keys()))):
        if attr == "parents" or attr == "properties":
            continue
        description += "{} differs:\n".format(attr)
        description += "old version: {}\n".format(
            olddiff[attr] if attr in olddiff else "not set")
        description += "new version: {}\n\n".format(
            newdiff[attr] if attr in newdiff else "not set")

    if len(olddiff["parents"]) > 0:
        description += ("Parents that are only in the old version:\n"
                        + ", ".join(olddiff["parents"]) + "\n")

    if len(newdiff["parents"]) > 0:
        description += ("Parents that are only in the new version:\n"
                        + ", ".join(olddiff["parents"]) + "\n")

    for prop in list(set(list(olddiff["properties"].keys())
                         + list(newdiff["properties"].keys()))):
        description += "property {} differs:\n".format(prop)

        if prop not in olddiff["properties"]:
            description += "it does not exist in the old version: \n"
        elif prop not in newdiff["properties"]:
            description += "it does not exist in the new version: \n"
        else:
            description += "old version: {}\n".format(
                olddiff["properties"][prop])
            description += "new version: {}\n\n".format(
                newdiff["properties"][prop])

    if description != "":
        description = ("## Difference between the old and the new "
                       "version of {}\n\n".format(name))+description

    return description


def apply_to_ids(entities, func):
    """ Apply a function to all ids.

    All ids means the ids of the entities themselves but also to all parents,
    properties and referenced entities.

    Parameters
    ----------
    entities : list of Entity
    func : function with one parameter.
    """

    for entity in entities:
        _apply_to_ids_of_entity(entity, func)


def _apply_to_ids_of_entity(entity, func):
    entity.id = func(entity.id)

    for par in entity.parents:
        par.id = func(par.id)

    for prop in entity.properties:
        prop.id = func(prop.id)
        isref = is_reference(prop.datatype)

        if isref:
            if isinstance(prop.value, list):
                prop.value = [func(el) for el in prop.value]
            else:
                if prop.value is not None:
                    prop.value = func(prop.value)


def resolve_reference(prop: Property):
    """resolves the value of a reference property

    The integer value is replaced with the entity object.
    If the property is not a reference, then the function returns without
    change.
    """

    if not prop.is_reference(server_retrieval=True):
        return

    if isinstance(prop.value, list):
        referenced = []

        for val in prop.value:
            if isinstance(val, int):
                referenced.append(retrieve_entity_with_id(val))
            else:
                referenced.append(val)
        prop.value = referenced
    else:
        if isinstance(prop.value, int):
            prop.value = retrieve_entity_with_id(prop.value)


def create_flat_list(ent_list: list[Entity], flat: list[Entity]):
    """
    Recursively adds all properties contained in entities from ent_list to
    the output list flat. Each element will only be added once to the list.

    TODO: Currently this function is also contained in newcrawler module crawl.
          We are planning to permanently move it to here.
    """
    for ent in ent_list:
        for p in ent.properties:
            # For lists append each element that is of type Entity to flat:
            if isinstance(p.value, list):
                for el in p.value:
                    if isinstance(el, Entity):
                        if el not in flat:
                            flat.append(el)
                        # TODO: move inside if block?
                        create_flat_list([el], flat)
            elif isinstance(p.value, Entity):
                if p.value not in flat:
                    flat.append(p.value)
                # TODO: move inside if block?
                create_flat_list([p.value], flat)


def _same_id_as_resolved_entity(this, that):
    """Checks whether ``this`` and ``that`` either are the same or whether one
    is an id and the other is a db.Entity with this id.

    """
    if isinstance(this, Entity) and not isinstance(that, Entity):
        # this is an Entity with an id, that is not
        return this.id is not None and this.id == that
    if not isinstance(this, Entity) and isinstance(that, Entity):
        return that.id is not None and that.id == this
    return this == that



# -*- coding: utf-8 -*-
#
# ** header v3.0
# This file is a part of the LinkAhead Project.
#
# Copyright (C) 2018 Research Group Biomedical Physics,
# Max-Planck-Institute for Dynamics and Self-Organization Göttingen
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
# ** end header
#

"""mising docstring."""
from lxml import etree
from multiprocessing import Lock
from uuid import uuid4
_uuid_lock = Lock()


def xml2str(xml):
    return etree.tostring(xml, pretty_print=True, encoding='unicode')


def uuid():
    exc = None
    ret = None
    try:
        _uuid_lock.acquire()
        ret = uuid4()
    except Exception as e:
        exc = e
    finally:
        _uuid_lock.release()
        if exc:
            raise exc
    return ret


def is_int(obj):
    try:
        int(obj)
        return True
    except ValueError:
        return False



# -*- coding: utf-8 -*-
#
# ** header v3.0
# This file is a part of the LinkAhead Project.
#
# Copyright (C) 2018 Research Group Biomedical Physics,
# Max-Planck-Institute for Dynamics and Self-Organization Göttingen
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
# ** end header
#
"""plain.

A CredentialsProvider which reads the password from the configuration
dict.
"""
from __future__ import absolute_import, unicode_literals, print_function
from .interface import CredentialsProvider, CredentialsAuthenticator


def get_authentication_provider():
    """get_authentication_provider.

    Return an authenticator which uses plain text username/password credentials.

    Returns
    -------
    CredentialsAuthenticator
        with a PlainTextCredentialsProvider as back-end.
    """
    return CredentialsAuthenticator(PlainTextCredentialsProvider())


class PlainTextCredentialsProvider(CredentialsProvider):
    """PlainTextCredentialsProvider.

    A class for storing username/password credentials as plain text strings.

    Methods
    -------
    configure

    Attributes
    ----------
    password
    username
    """

    def __init__(self):
        super(PlainTextCredentialsProvider, self).__init__()
        self._password = None
        self._username = None

    def configure(self, **config):
        """configure.

        Parameters
        ----------
        **config
            Keyword arguments containing at least keywords "username" and "password".

        Returns
        -------
        None
        """
        if "password" in config:
            self._password = config["password"]
        if "username" in config:
            self._username = config["username"]

    @property
    def password(self):
        return self._password

    @property
    def username(self):
        return self._username



# -*- coding: utf-8 -*-
#
# ** header v3.0
# This file is a part of the LinkAhead Project.
#
# Copyright (C) 2018 Research Group Biomedical Physics,
# Max-Planck-Institute for Dynamics and Self-Organization Göttingen
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
# ** end header
#

"""Utilities for work with PlantUML.

PlantUML (http://plantuml.com) is a converter from a simple
descriptive language to graphviz diagrams.

To convert the output, you can write it into FILENAME.pu and then
convert it with:

plantuml FILENAME.pu -> FILENAME.png
"""

import os
import shutil

import linkahead as db
from linkahead.common.datatype import is_reference, get_referenced_recordtype

from typing import List, Optional

import tempfile

REFERENCE = "REFERENCE"


def get_description(description_str):
    """Extract and format a description string from a record type or property.

    Parameters
    ----------
    description_str : str
                      The description string that is going to be formatted.

    Returns
    -------
    str
       The reformatted description ending in a line break.
    """
    words = description_str.split()
    lines = []
    lines.append("")

    for w in words:
        if len(lines[-1] + w) > 60:
            lines.append("")

        if len(lines[-1]) > 0:
            lines[-1] += " "
        lines[-1] += w
    description = "\n".join(lines)

    return description + "\n"


class Grouped(object):
    def __init__(self, name, parents):
        self.name = name
        self.parents = parents

    def get_parents(self):
        return self.parents


def recordtypes_to_plantuml_string(iterable,
                                   add_properties: bool = True,
                                   add_recordtypes: bool = True,
                                   add_legend: bool = True,
                                   no_shadow: bool = False,
                                   style: str = "default"):
    """Converts RecordTypes into a string for PlantUML.

    This function obtains an iterable and returns a string which can
    be input into PlantUML for a representation of all RecordTypes in
    the iterable.

    Current options for style
    -------------------------

    "default" - Standard rectangles with uml class circle and methods section
    "salexan" - Round rectangles, hide circle and methods section

    Current limitations
    -------------------

    - It is inherently hard to detect if an element should be rendered
      as a class/RecordType or not.  Currently it is rendered if
      either the "type" attribute is None or
      type(element) == RecordType.
    - Inheritance of Properties is not rendered nicely at the moment.

    Parameters
    ----------
    iterable: iterable of linkahead.Entity
      The objects to be rendered with plantuml.

    no_shadow : bool, optional
      If true, tell plantuml to use a skin without blurred shadows.


    Returns
    -------
    out : str
      The plantuml string for the given container.
    """

    # TODO: This function needs a review of python type hints.

    classes = [el for el in iterable
               if isinstance(el, db.RecordType)]
    dependencies = {}
    inheritances = {}
    properties = [p for p in iterable if isinstance(p, db.Property)]
    grouped = [g for g in iterable if isinstance(g, Grouped)]

    def _add_properties(c, importance=None):
        result = ""

        for p in c.get_properties():
            if importance is None or c.get_properties().get_importance(p) == importance:
                if importance is not None and len(result) == 0:
                    result += ".." + importance.lower() + "..\n"
                name = p.name
                p_type = p.datatype

                if p_type is None:
                    # get type from properties

                    for p2 in properties:
                        if p2.name == p.name:
                            p_type = p2.datatype

                if p_type is None:
                    # is reference?

                    for p2 in classes:
                        if p2.name == p.name:
                            p_type = p2

                if isinstance(p_type, db.Entity):
                    p_type = p_type.name
                    dependencies[c].append(p_type)
                elif p_type is not None:
                    for c2 in classes:
                        if c2.name == p_type or db.LIST(c2.name) == p_type:
                            dependencies[c].append(c2.name)
                result += '  {name} ({type})\n'.format(
                    name=name, type=p_type)

        return result

    result = "@startuml\n\n"

    if no_shadow:
        result += "skinparam shadowing false\n"

    if style == "default":
        result += "skinparam classAttributeIconSize 0\n"
    elif style == "salexan":
        result += """skinparam roundcorner 20\n
skinparam boxpadding 20\n
\n
hide methods\n
hide circle\n
"""
    else:
        raise ValueError("Unknown style.")

    if add_properties:
        result += "package Properties #DDDDDD {\n"
        for p in properties:
            inheritances[p] = p.get_parents()
            dependencies[p] = []

            result += "class \"{klass}\" << (P,#008800) >> {{\n".format(klass=p.name)

            if p.description is not None:
                result += get_description(p.description)
            result += "\n..\n"

            if isinstance(p.datatype, str):
                result += "datatype: " + p.datatype + "\n"
            elif isinstance(p.datatype, db.Entity):
                result += "datatype: " + p.datatype.name + "\n"
            else:
                result += "datatype: " + str(p.datatype) + "\n"
            result += "}\n\n"
        result += "}\n\n"

    if add_recordtypes:
        result += "package RecordTypes #DDDDDD {\n"

        for c in classes:
            inheritances[c] = c.get_parents()
            dependencies[c] = []
            result += "class \"{klass}\" << (C,#FF1111) >> {{\n".format(klass=c.name)

            if c.description is not None:
                result += get_description(c.description)

            props = ""
            props += _add_properties(c, importance=db.FIX)
            props += _add_properties(c, importance=db.OBLIGATORY)
            props += _add_properties(c, importance=db.RECOMMENDED)
            props += _add_properties(c, importance=db.SUGGESTED)

            if len(props) > 0:
                result += "__Properties__\n" + props
            else:
                result += "\n..\n"
            result += "}\n\n"

        for g in grouped:
            inheritances[g] = g.get_parents()
            result += "class \"{klass}\" << (G,#0000FF) >> {{\n".format(klass=g.name)
        result += "}\n\n"

        for c, parents in inheritances.items():
            for par in parents:
                result += "\"{par}\" <|-- \"{klass}\"\n".format(
                    klass=c.name, par=par.name)

        for c, deps in dependencies.items():
            for dep in deps:
                result += "\"{klass}\" *-- \"{dep}\"\n".format(
                    klass=c.name, dep=dep)

    if add_legend:
        result += """

package \"B is a subtype of A\" <<Rectangle>> {
 A <|-right- B
 note  "This determines what you find when you query for the RecordType.\\n'FIND RECORD A' will provide Records which have a parent\\nA or B, while 'FIND RECORD B' will provide only Records which have a parent B." as N1
}
"""
        result += """

package \"The property P references an instance of D\" <<Rectangle>> {
 class C {
    P(D)
 }
 C *-right- D
 note  "Employ this when searching for C: 'FIND RECORD C WITH D'\\nOr if the value of D is a Record: 'FIND RECORD C WHICH REFERENCES D' is possible.\\nEmploying this while searching for D: 'FIND RECORD D WHICH IS REFERENCED BY C" as N2
}

"""

    result += "\n@enduml\n"

    return result


def retrieve_substructure(start_record_types, depth, result_id_set=None, result_container=None, cleanup=True):
    """Recursively retrieves LinkAhead record types and properties, starting
    from given initial types up to a specific depth.

    Parameters
    ----------
    start_record_types : Iterable[db.Entity]
                         Iterable with the entities to be displayed. Starting from these
                         entities more entities will be retrieved.
    depth : int
            The maximum depth up to which to retriev sub entities.
    result_id_set : set[int]
                    Used by recursion. Filled with already visited ids.
    result_container : db.Container
                       Used by recursion. Filled with already visited entities.
    cleanup : bool
              Used by recursion. If True return the resulting result_container.
              Don't return anything otherwise.

    Returns
    -------
    db.Container
                A container containing all the retrieved entites
                or None if cleanup is False.
    """
    # Initialize the id set and result container for level zero recursion depth:
    if result_id_set is None:
        result_id_set = set()
    if result_container is None:
        result_container = db.Container()

    for entity in start_record_types:
        entity.retrieve()
        if entity.id not in result_id_set:
            result_container.append(entity)
        result_id_set.add(entity.id)
        for prop in entity.properties:
            if (is_reference(prop.datatype) and prop.datatype != db.FILE and depth > 0):
                rt = db.RecordType(
                    name=get_referenced_recordtype(prop.datatype)).retrieve()
                retrieve_substructure([rt], depth-1, result_id_set,
                                      result_container, False)
            # TODO: clean up this hack
            # TODO: make it also work for files
            if is_reference(prop.datatype) and prop.value is not None:
                r = db.Record(id=prop.value).retrieve()
                retrieve_substructure([r], depth-1, result_id_set, result_container, False)
                if r.id not in result_id_set:
                    result_container.append(r)
                    result_id_set.add(r.id)

            if prop.id not in result_id_set:
                result_container.append(prop)
                result_id_set.add(prop.id)

        for parent in entity.parents:
            rt = db.RecordType(id=parent.id).retrieve()
            if parent.id not in result_id_set:
                result_container.append(rt)
            result_id_set.add(parent.id)
            if depth > 0:
                retrieve_substructure([rt], depth-1, result_id_set,
                                      result_container, False)

    if cleanup:
        return result_container
    return None


def to_graphics(recordtypes: List[db.Entity], filename: str,
                output_dirname: Optional[str] = None,
                formats: List[str] = ["tsvg"],
                silent: bool = True,
                add_properties: bool = True,
                add_recordtypes: bool = True,
                add_legend: bool = True,
                no_shadow: bool = False,
                style: str = "default"):
    """Calls recordtypes_to_plantuml_string(), saves result to file and
    creates an svg image

    plantuml needs to be installed.

    Parameters
    ----------
    recordtypes : Iterable[db.Entity]
                  Iterable with the entities to be displayed.
    filename : str
               filename of the image without the extension(e.g. data_structure;
               also without the preceeding path.
               data_structure.pu and data_structure.svg will be created.)
    output_dirname : str
                     the destination directory for the resulting images as defined by the "-o"
                     option by plantuml
                     default is to use current working dir
    formats : List[str]
              list of target formats as defined by the -t"..." options by plantuml, e.g. "tsvg"
    silent : bool
             Don't output messages.
    no_shadow : bool, optional
      If true, tell plantuml to use a skin without blurred shadows.
    """
    pu = recordtypes_to_plantuml_string(iterable=recordtypes,
                                        add_properties=add_properties,
                                        add_recordtypes=add_recordtypes,
                                        add_legend=add_legend,
                                        no_shadow=no_shadow,
                                        style=style)

    if output_dirname is None:
        output_dirname = os.getcwd()

    allowed_formats = [
        "tpng", "tsvg", "teps", "tpdf", "tvdx", "txmi",
        "tscxml", "thtml", "ttxt", "tutxt", "tlatex", "tlatex:nopreamble"]

    with tempfile.TemporaryDirectory() as td:

        pu_filename = os.path.join(td, filename + ".pu")
        with open(pu_filename, "w") as pu_file:
            pu_file.write(pu)

        for format in formats:
            extension = format[1:]
            if ":" in extension:
                extension = extension[:extension.index(":")]

            if format not in allowed_formats:
                raise RuntimeError("Format not allowed.")
            cmd = "plantuml -{} {}".format(format, pu_filename)
            if not silent:
                print("Executing:", cmd)

            if os.system(cmd) != 0:  # TODO: replace with subprocess.run
                raise Exception("An error occured during the execution of "
                                "plantuml when using the format {}. "
                                "Is plantuml installed? "
                                "You might want to try a different format.".format(format))
            # copy only the final product into the target directory
            shutil.copy(os.path.join(td, filename + "." + extension),
                        output_dirname)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
#
#
"""linkahead"""
import os
import re
import subprocess
import sys

from setuptools import find_packages, setup

########################################################################
# The following code is largely based on code in numpy
########################################################################
#
# Copyright (c) 2005-2019, NumPy Developers.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#    * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#
#    * Redistributions in binary form must reproduce the above
#       copyright notice, this list of conditions and the following
#       disclaimer in the documentation and/or other materials provided
#       with the distribution.
#
#    * Neither the name of the NumPy Developers nor the names of any
#       contributors may be used to endorse or promote products derived
#       from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
########################################################################

ISRELEASED = True
MAJOR = 0
MINOR = 15
MICRO = 1
# Do not tag as pre-release until this commit
# https://github.com/pypa/packaging/pull/515
# has made it into a release. Probably we should wait for pypa/packaging>=21.4
# https://github.com/pypa/packaging/releases
PRE = ""  # "dev"  # e.g. rc0, alpha.1, 0.beta-23

if PRE:
    VERSION = "{}.{}.{}-{}".format(MAJOR, MINOR, MICRO, PRE)
else:
    VERSION = "{}.{}.{}".format(MAJOR, MINOR, MICRO)


# Return the git revision as a string
def git_version():
    def _minimal_ext_cmd(cmd):
        # construct minimal environment
        env = {}

        for k in ['SYSTEMROOT', 'PATH', 'HOME']:
            v = os.environ.get(k)

            if v is not None:
                env[k] = v
        # LANGUAGE is used on win32
        env['LANGUAGE'] = 'C'
        env['LANG'] = 'C'
        env['LC_ALL'] = 'C'
        out = subprocess.check_output(cmd, stderr=subprocess.STDOUT, env=env)

        return out

    try:
        out = _minimal_ext_cmd(['git', 'rev-parse', 'HEAD'])
        GIT_REVISION = out.strip().decode('ascii')
    except (subprocess.SubprocessError, OSError):
        GIT_REVISION = "Unknown"

    return GIT_REVISION


def get_version_info():
    # Adding the git rev number needs to be done inside write_version_py(),
    # otherwise the import of linkahead.version messes up the build under
    # Python 3.
    FULLVERSION = VERSION

    # Magic which is only really needed in the pipelines. Therefore: a lot of dark pipeline magic.
    GIT_REVISION = "Unknown"
    if os.path.exists('.git'):
        GIT_REVISION = git_version()
    elif os.path.exists('linkahead_pylib_commit'):
        with open('linkahead_pylib_commit', 'r') as f:
            GIT_REVISION = f.read().strip()
    elif os.path.exists('src/linkahead/version.py'):
        # must be a source distribution, use existing version file
        with open('src/linkahead/version.py') as fi:
            rev_pattern = re.compile(r"^git_revision = '(?P<rev>.*)'$")
            for line in fi.readlines():
                match = rev_pattern.match(line)
                if match is not None:
                    GIT_REVISION = match.group('rev')
                    break

    if not ISRELEASED:
        FULLVERSION += '.dev0+' + GIT_REVISION[:7]

    return FULLVERSION, GIT_REVISION


def write_version_py(filename='src/linkahead/version.py'):
    cnt = """
# THIS FILE IS GENERATED FROM linkahead SETUP.PY
#
short_version = '%(version)s'
version = '%(version)s'
full_version = '%(full_version)s'
git_revision = '%(git_revision)s'
release = %(isrelease)s

if not release:
    version = full_version
"""
    FULLVERSION, GIT_REVISION = get_version_info()

    a = open(filename, 'w')
    try:
        a.write(cnt % {'version': VERSION,
                       'full_version': FULLVERSION,
                       'git_revision': GIT_REVISION,
                       'isrelease': str(ISRELEASED)})
    finally:
        a.close()


def setup_package():
    # load README
    with open("README.md", "r") as fh:
        long_description = fh.read()

    src_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), "src")
    sys.path.insert(0, src_path)

    # Rewrite the version file everytime
    write_version_py()

    if 'PKGNAME' in os.environ and os.environ['PKGNAME'] == 'caosdb':
        pname = 'caosdb'
        pdesc = 'Deprecated! Please install linkahead.'
    else:
        pname = 'linkahead'
        pdesc = 'Python Interface for LinkAhead'
    metadata = dict(
        name=pname,
        version=get_version_info()[0],
        description=pdesc,
        long_description=long_description,
        long_description_content_type="text/markdown",
        author='Timm Fitschen',
        author_email='t.fitschen@indiscale.com',
        url='https://www.linkahead.org',
        license="AGPLv3+",
        classifiers=[
            "Programming Language :: Python :: 3",
            "License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)",
            "Operating System :: OS Independent",
            "Topic :: Database",
            "Topic :: Scientific/Engineering :: Information Analysis",
        ],
        packages=find_packages('src'),
        python_requires='>=3.8',
        package_dir={'': 'src'},
        install_requires=['lxml>=4.6.3',
                          "requests[socks]>=2.26",
                          "python-dateutil>=2.8.2",
                          'PyYAML>=5.4.1',
                          'future',
                          ],
        extras_require={'keyring': ['keyring>=13.0.0'],
                        'jsonschema': ['jsonschema>=4.4.0']},
        setup_requires=["pytest-runner>=2.0,<3dev"],
        tests_require=["pytest", "pytest-cov", "coverage>=4.4.2",
                       "jsonschema>=4.4.0"],
        package_data={
            'linkahead': ['py.typed', 'cert/indiscale.ca.crt', 'schema-pycaosdb-ini.yml'],
        },
        scripts=[
            "src/linkahead/utils/caosdb_admin.py",
            "src/linkahead/utils/linkahead_admin.py"
        ]
    )
    try:
        setup(**metadata)
    finally:
        del sys.path[0]
    return


if __name__ == '__main__':
    setup_package()



# -*- coding: utf-8 -*-
#
# This file is a part of the LinkAhead Project.
#
# Copyright (C) 2023 IndiScale GmbH <info@indiscale.com>
# Copyright (C) 2023 Henrik tom Wörden <h.tomwoerden@indiscale.com>
# Copyright (C) 2023 Daniel Hornung <d.hornung@indiscale.com>
# Copyright (C) 2024 Indiscale GmbH <info@indiscale.com>
# Copyright (C) 2024 Joscha Schmiedt <joscha@schmiedt.dev>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

"""
This module provides some cached versions of functions that retrieve Entities from a remote server.

See also
========

- ``cache_initialize(...)`` : Re-initialize the cache.
- ``cache_clear()`` : Clear the cache.
- ``cached_query(query)`` : A cached version of ``execute_query(query)``.
- ``cached_get_entity_by(...)`` : Get an Entity by name, id, ...
"""

from __future__ import annotations
from enum import Enum
from functools import lru_cache
from typing import Any, Optional, Union

from .exceptions import EmptyUniqueQueryError, QueryNotUniqueError
from .utils import get_entity
from .common.models import execute_query, Entity, Container


# roughly 1GB for typical entity sizes
DEFAULT_SIZE = 33333

# This dict cache is solely for filling the real cache manually (e.g. to reuse older query results)
_DUMMY_CACHE: dict[Union[str, int], Any] = {}


class AccessType(Enum):
    """Different access types for cached queries.  Needed for filling the cache manually with
:func:`cache_fill` .

    """
    QUERY = 1
    PATH = 2
    EID = 3
    NAME = 4


def cached_get_entity_by(eid: Union[str, int, None] = None,
                         name: Optional[str] = None,
                         path: Optional[str] = None,
                         query: Optional[str] = None) -> Union[Entity, tuple[None]]:
    """Return a single entity that is identified uniquely by one argument.

You must supply exactly one argument.

If a query phrase is given, the result must be unique.  If this is not what you need, use
:func:`cached_query` instead.

    """
    count = 0
    if eid is not None:
        count += 1
    if name is not None:
        count += 1
    if path is not None:
        count += 1
    if query is not None:
        count += 1
    if count != 1:
        raise ValueError("You must supply exactly one argument.")

    result = (None, )
    if eid is not None:
        result = _cached_access(AccessType.EID, eid, unique=True)
    if name is not None:
        result = _cached_access(AccessType.NAME, name, unique=True)
    if path is not None:
        result = _cached_access(AccessType.PATH, path, unique=True)
    if query is not None:
        result = _cached_access(AccessType.QUERY, query, unique=True)

    if result != (None, ):
        if isinstance(result, (QueryNotUniqueError, EmptyUniqueQueryError)):
            raise result
        return result

    raise RuntimeError("This line should never be reached.")


def cached_query(query_string: str) -> Container:
    """A cached version of :func:`linkahead.execute_query<linkahead.common.models.execute_query>`.

All additional arguments are at their default values.

    """
    result = _cached_access(AccessType.QUERY, query_string, unique=False)
    if isinstance(result, (QueryNotUniqueError, EmptyUniqueQueryError)):
        raise result
    return result


@ lru_cache(maxsize=DEFAULT_SIZE)
def _cached_access(kind: AccessType, value: Union[str, int], unique: bool = True):
    # This is the function that is actually cached.
    # Due to the arguments, the cache has kind of separate sections for cached_query and
    # cached_get_entity_by with the different AccessTypes. However, there is only one cache size.

    # The dummy dict cache is only for filling the cache manually, it is deleted afterwards.
    if value in _DUMMY_CACHE:
        return _DUMMY_CACHE[value]

    try:
        if kind == AccessType.QUERY:
            if not isinstance(value, str):
                raise TypeError(
                    f"If AccessType is QUERY, value must be a string, not {type(value)}.")
            return execute_query(value, unique=unique)
        if kind == AccessType.NAME:
            if not isinstance(value, str):
                raise TypeError(
                    f"If AccessType is NAME, value must be a string, not {type(value)}.")
            return get_entity.get_entity_by_name(value)
        if kind == AccessType.EID:
            if not isinstance(value, (str, int)):
                raise TypeError(
                    f"If AccessType is EID, value must be a string or int, not {type(value)}.")
            return get_entity.get_entity_by_id(value)
        if kind == AccessType.PATH:
            if not isinstance(value, str):
                raise TypeError(
                    f"If AccessType is PATH, value must be a string, not {type(value)}.")
            return get_entity.get_entity_by_path(value)
    except (QueryNotUniqueError, EmptyUniqueQueryError) as exc:
        return exc

    raise ValueError(f"Unknown AccessType: {kind}")


def cache_clear() -> None:
    """Empty the cache that is used by `cached_query` and `cached_get_entity_by`."""
    _cached_access.cache_clear()


def cache_info():
    """Return info about the cache that is used by `cached_query` and `cached_get_entity_by`.

Returns
-------

out: named tuple
  See the standard library :func:`functools.lru_cache` for details."""
    return _cached_access.cache_info()


def cache_initialize(maxsize: int = DEFAULT_SIZE) -> None:
    """Create a new cache with the given size for `cached_query` and `cached_get_entity_by`.

    This implies a call of :func:`cache_clear`, the old cache is emptied.

    """
    cache_clear()
    global _cached_access
    _cached_access = lru_cache(maxsize=maxsize)(_cached_access.__wrapped__)


def cache_fill(items: dict[Union[str, int], Any],
               kind: AccessType = AccessType.EID,
               unique: bool = True) -> None:
    """Add entries to the cache manually.

    This allows to fill the cache without actually submitting queries.  Note that this does not
    overwrite existing entries with the same keys.

Parameters
----------

items: dict
  A dictionary with the entries to go into the cache.  The keys must be compatible with the
  AccessType given in ``kind``

kind: AccessType, optional
  The AccessType, for example ID, name, path or query.

unique: bool, optional
  If True, fills the cache for :func:`cached_get_entity_by`, presumably with
  :class:`linkahead.Entity<linkahead.common.models.Entity>` objects.  If False, the cache should be
  filled with :class:`linkahead.Container<linkahead.common.models.Container>` objects, for use with
  :func:`cached_query`.

    """

    if kind == AccessType.QUERY:
        assert all(isinstance(key, str) for key in items.keys()), "Keys must be strings."
    elif kind == AccessType.NAME:
        assert all(isinstance(key, str) for key in items.keys()), "Keys must be strings."
    elif kind == AccessType.EID:
        assert all(isinstance(key, (str, int))
                   for key in items.keys()), "Keys must be strings or integers."
    elif kind == AccessType.PATH:
        assert all(isinstance(key, str) for key in items.keys()), "Keys must be strings."
    else:
        raise ValueError(f"Unknown AccessType: {kind}")

    # 1. add the given items to the corresponding dummy dict cache
    _DUMMY_CACHE.update(items)

    # 2. call the cache function with each key (this only results in a dict look up)
    for key in items.keys():
        _cached_access(kind, key, unique=unique)

    # 3. empty the dummy dict cache again
    _DUMMY_CACHE.clear()



# -*- coding: utf-8 -*-
#
# This file is a part of the LinkAhead Project.
#
# Copyright (C) 2024 Henrik tom Wörden <h.tomwoerden@indiscale.com>
# Copyright (C) 2024 IndiScale GmbH <info@indiscale.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

import warnings


def escape_squoted_text(text: str) -> str:
    r"""Return an escaped version of the argument.

    The characters ``\``, ``*`` and ``'`` need to be escaped if used in single quoted
    expressions in the query language.

    This function returns the given string where the characters ``\``, ``'`` and ``*`` are
    escaped by a ``\`` (backslash character).

    Parameters
    ----------
    text : str
        The text to be escaped.

    Returns
    -------
    out : str
        The escaped text.
    """
    return text.replace("\\", r"\\").replace("'", r"\'").replace("*", r"\*")


def escape_dquoted_text(text: str) -> str:
    r"""Return an escaped version of the argument.

    The characters ``\``, ``*`` and ``"`` need to be escaped if used in double quoted
    expressions in the query language.

    This function returns the given string where the characters ``\``, ``"`` and ``*`` are
    escaped by a ``\`` (backslash character).

    Parameters
    ----------
    text : str
        The text to be escaped.

    Returns
    -------
    out : str
        The escaped text.
    """
    return text.replace("\\", r"\\").replace('"', r"\"").replace("*", r"\*")


def escape_quoted_text(text: str) -> str:
    """
    Please use escape_squoted_text or escape_dquoted_text instead of this function.
    """
    warnings.warn("Please use escape_squoted_text or escape_dquoted_text", DeprecationWarning)
    return escape_squoted_text(text)
